#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Modified version from QTPyLib: Quantitative Trading Python Library
# https://github.com/ranaroussi/qtpylib
# Copyright 2016-2018 Ran Aroussi
#
# Modified by vin8tech
# Copyright 2019-2021 vin8tech, vinay patil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import argparse
import atexit
import json
import logging
import os
import pickle

import sys
import tempfile
import time
import glob
import subprocess

from datetime import datetime
from abc import ABCMeta
from math import ceil

import zmq
import pandas as pd
from dateutil.parser import parse as parse_date

from numpy import (
    isnan as np_isnan,
    nan as np_nan,
    int64 as np_int64
)

from kinetick import (
    path,
)
from kinetick.utils import utils, asynctools
from mongoengine import connect as mongo_connect, NotUniqueError, Q
from kinetick.enums import Timeframes, COMMON_TYPES
from kinetick.lib.brokers import Webull
from kinetick.models import Tick, OHLC
from kinetick.utils.utils import (
    read_single_argv,
    is_number,
    create_logger, force_options_columns, datetime64_to_datetime, wb_lookback_str, chmod, resample,
    create_continuous_contract, gen_symbol_group, gen_asset_class
)

# =============================================
# Configure logging
create_logger(__name__, os.getenv('LOGLEVEL') or logging.INFO)
logging.getLogger('zmq').setLevel(os.getenv('LOGLEVEL') or logging.INFO)

# =============================================
# set up threading pool
__threads__ = read_single_argv("--threads")
__threads__ = __threads__ if is_number(__threads__) else None
asynctools.multitasking.createPool(__name__, __threads__)

# =============================================

cash_ticks = {}


class Blotter():
    """Broker class initilizer

    :Optional:

        name : string
            name of the blotter (used by other modules)
        symbols : str
            contracts CSV database (default: ./symbols.csv)
        zmqport : str
            ZeroMQ Port to use (default: 12345)
        zmqtopic : str
            ZeroMQ string to use (default: _kinetick_BLOTTERNAME_)
        orderbook : str
            Get Order Book (Market Depth) data (default: False)
        dbhost : str
            MySQL server hostname (default: localhost)
        dbport : str
            MySQL server port (default: 3306)
        dbname : str
            MySQL server database (default: kinetick)
        dbuser : str
            MySQL server username (default: root)
        dbpass : str
            MySQL server password (default: none)
        dbskip : str
            Skip MySQL logging (default: True)
    """

    __metaclass__ = ABCMeta

    def __init__(self, name=None, symbols="symbols.csv",
                 dbhost="localhost", dbport="27017", dbname="kinetick",
                 dbuser="root", dbpass="", dbskip=True, orderbook=False,
                 zmqport="12345", zmqtopic=None, **kwargs):

        # whats my name?
        self.name = str(self.__class__).split('.')[-1].split("'")[0].lower()
        if name is not None:
            self.name = name

        # initialize class logger
        self.log_blotter = logging.getLogger(__name__)

        # do not act on first tick (timezone is incorrect)
        self.first_tick = True

        self._bars = pd.DataFrame(
            columns=['open', 'high', 'low', 'close', 'volume'])
        self._bars.index.names = ['datetime']
        self._bars.index = pd.to_datetime(self._bars.index, utc=True)
        # self._bars.index = self._bars.index.tz_convert(settings['timezone'])
        self._bars = {"~": self._bars}

        self._raw_bars = pd.DataFrame(columns=['last', 'volume'])
        self._raw_bars.index.names = ['datetime']
        self._raw_bars.index = pd.to_datetime(self._raw_bars.index, utc=True)
        self._raw_bars = {"~": self._raw_bars}

        # global objects
        self.db_connection = None
        self.context = None
        self.socket = None
        self.connection = None  # stream provider

        self.symbol_ids = {}  # cache
        self.cash_ticks = cash_ticks  # outside cache
        self.rtvolume = set()  # has RTVOLUME?

        # override args with any (non-default) command-line args
        self.args = {arg: val for arg, val in locals().items(
        ) if arg not in ('__class__', 'self', 'kwargs')}
        self.args.update(kwargs)
        self.args.update(self.load_cli_args())

        # -------------------------------
        # work default values
        # -------------------------------
        if zmqtopic is None:
            zmqtopic = "_kinetick_" + str(self.name.lower()) + "_"
            if self.args['zmqtopic'] is None:
                self.args['zmqtopic'] = zmqtopic

        # if no path given for symbols' csv, use same dir
        if symbols == "symbols.csv":
            symbols = path['caller'] + '/' + symbols
        # -------------------------------

        # read cached args to detect duplicate blotters
        self.duplicate_run = False
        self.cached_args = {}
        self.args_cache_file = "%s/%s.kinetick" % (
            tempfile.gettempdir(), self.name)
        if os.path.exists(self.args_cache_file):
            self.cached_args = self._read_cached_args()

        # don't display connection errors on ctrl+c
        self.quitting = False

        # do stuff on exit
        atexit.register(self._on_exit)

        # track historical data download status
        self.backfilled = False
        self.backfilled_symbols = []
        self.backfill_resolution = Timeframes.MINUTE_1  # default to 1 min

        # be aware of thread count
        self.threads = asynctools.multitasking.getPool(__name__)['threads']

    # -------------------------------------------
    def _on_exit(self, terminate=True):
        if "as_client" in self.args:
            return

        self.log_blotter.info("Blotter stopped...")

        if hasattr(self, 'wb'):
            self.log_blotter.info("Cancel market data...")
            self.connection.cancelMarketDataSubscription()

            self.log_blotter.info("Disconnecting...")
            self.connection.disconnect()

        if not self.duplicate_run:
            self.log_blotter.info("Deleting runtime args...")
            self._remove_cached_args()

        if not self.args['dbskip']:
            self.log_blotter.info("Disconnecting from DB...")
            try:
                self.db_connection.close()
            except Exception as e:
                pass

        if terminate:
            sys.exit(0)

    # -------------------------------------------
    @staticmethod
    def _detect_running_blotter(name):
        return name

    # -------------------------------------------
    @staticmethod
    def _blotter_file_running():
        try:
            # not sure how this works on windows...
            command = 'pgrep -f ' + sys.argv[0]
            process = subprocess.Popen(
                command, shell=True, stdout=subprocess.PIPE)
            stdout_list = process.communicate()[0].decode('utf-8').split("\n")
            stdout_list = list(filter(None, stdout_list))
            return len(stdout_list) > 0
        except Exception as e:
            return False

    # -------------------------------------------
    def _check_unique_blotter(self):
        if os.path.exists(self.args_cache_file):
            # temp file found - check if really running
            # or if this file wasn't deleted due to crash
            if not self._blotter_file_running():
                # print("REMOVING OLD TEMP")
                self._remove_cached_args()
            else:
                self.duplicate_run = True
                self.log_blotter.error("Blotter is already running...")
                # sys.exit(1)

        self._write_cached_args()

    # -------------------------------------------
    def _remove_cached_args(self):
        if os.path.exists(self.args_cache_file):
            os.remove(self.args_cache_file)

    def _read_cached_args(self):
        if os.path.exists(self.args_cache_file):
            return pickle.load(open(self.args_cache_file, "rb"))
        return {}

    def _write_cached_args(self):
        pickle.dump(self.args, open(self.args_cache_file, "wb"))
        utils.chmod(self.args_cache_file)

    # -------------------------------------------
    def load_cli_args(self):
        def to_boolean(s):
            if s:
                return s.lower() in ['true', '1', 't', 'y', 'yes']
            else:
                return None

        parser = argparse.ArgumentParser(
            description='Kinetick Blotter',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)

        parser.add_argument('--symbols', default=os.getenv("symbols") or self.args['symbols'],
                            help='contracts CSV database', required=False)
        parser.add_argument('--zmqport', default=os.getenv("zmqport") or self.args['zmqport'],
                            help='ZeroMQ Port to use', required=False)
        parser.add_argument('--zmqtopic', default=os.getenv("zmqtopic") or self.args['zmqtopic'],
                            help='ZeroMQ Port to use', required=False)
        parser.add_argument('--orderbook', action='store_true',
                            default=to_boolean(os.getenv("orderbook")) or self.args['orderbook'],
                            help='Get Order Book (Market Depth) data',
                            required=False)
        parser.add_argument('--dbhost', default=os.getenv("dbhost") or self.args['dbhost'],
                            help='MySQL server hostname', required=False)
        parser.add_argument('--dbport', default=os.getenv("dbport") or self.args['dbport'],
                            help='MySQL server port', required=False)
        parser.add_argument('--dbname', default=os.getenv("dbname") or self.args['dbname'],
                            help='MySQL server database', required=False)
        parser.add_argument('--dbuser', default=os.getenv("dbuser") or self.args['dbuser'],
                            help='MySQL server username', required=False)
        parser.add_argument('--dbpass', default=os.getenv("dbpass") or self.args['dbpass'],
                            help='MySQL server password', required=False)
        dbskip = to_boolean(os.getenv("dbskip")) if os.getenv("dbskip") is not None else self.args['dbskip']
        parser.add_argument('--dbskip', default=dbskip,
                            required=False, help='Skip DB logging (flag)',
                            action='store_true')

        # only return non-default cmd line args
        # (meaning only those actually given)
        cmd_args, _ = parser.parse_known_args()
        args = {arg: val for arg, val in vars(
            cmd_args).items()}
        return args

    # -------------------------------------------
    def callbacks(self, caller, msg, **kwargs):
        # self.log_blotter.debug("caller: [%s]", caller)
        # self.log_blotter.debug("Message Received: %s", msg)

        if caller == "handleConnectionClosed":
            self.log_blotter.info("Lost connection to Broker...")
            # let docker handle the restarts
            self._on_exit(terminate=True)
            # self.run()

        elif caller == "handleHistoricalData":
            self.on_ohlc_received(msg, kwargs)

        # elif caller == "handleTickString":
        #     self.on_tick_string_received(msg['tickerId'], kwargs)

        elif caller == "handleMarketQuote":
            self.on_quote_received(kwargs['tickerId'])

        elif caller == "handleTickPrice" or caller == "handleTickSize" or caller == "handleTickString":
            tickerId = msg['tickerId']
            self.on_tick_string_received(tickerId, kwargs)

        elif caller in "handleTickOptionComputation":
            self.on_option_computation_received(msg['tickerId'])

        elif caller == "handleMarketDepth":
            self.on_orderbook_received(msg['tickerId'])

        elif caller == "handleError":
            # don't display connection errors on ctrl+c
            if self.quitting:
                return

    # -------------------------------------------
    def on_ohlc_received(self, msg, kwargs):
        symbol = self.connection.tickerSymbol(kwargs['tickerId'])

        # print(msg)
        # msg.index.names = ['datetime']
        msg['datetime'] = msg.index
        data = msg.to_dict(orient='records')

        for row in data:
            params = {"tickerId": str(kwargs['tickerId']), "symbol": symbol,
                      # "symbol_group": utils.gen_symbol_group(symbol), "asset_class": utils.gen_asset_class(
                      # symbol),
                      "datetime": utils.datetime_to_timezone(
                          parse_date(str(row['datetime'])), tz="UTC"
                      ).strftime("%Y-%m-%d %H:%M:%S"), "open": utils.to_decimal(row['open']),
                      "high": utils.to_decimal(row['high']), "low": utils.to_decimal(row['low']),
                      "close": utils.to_decimal(row['close']), "volume": int(row['volume']),
                      "vwap": utils.to_decimal(row['vwap']), "resolution": self.backfill_resolution}

            ohlc = OHLC(**params)
            # store in db
            try:
                self.log2db(data=ohlc)
            except NotUniqueError:
                pass
                # self.log_blotter.info(f'duplicate row skipped. {row}')

        if kwargs["completed"]:
            self.backfilled_symbols.append(symbol)
            tickers = set(
                {v: k for k, v in self.connection.tickerIds.items() if v.upper() != "SYMBOL"}.keys())
            if tickers == set(self.backfilled_symbols):
                self.backfilled = True
        print(".")

    # -------------------------------------------
    @asynctools.multitasking.task
    def on_tick_string_received(self, tickerId, kwargs):

        # kwargs is empty
        # if not kwargs:
        #     return

        data = None
        symbol = self.connection.tickerSymbol(tickerId)

        # for instruments that receive RTVOLUME events
        if "tick" in kwargs:
            self.rtvolume.add(symbol)
            data = {
                # available data from ib
                "symbol": symbol,
                "symbol_group": utils.gen_symbol_group(symbol),  # ES_F, ...
                "asset_class": utils.gen_asset_class(symbol),
                "timestamp": kwargs['tick']['time'],
                "last": utils.to_decimal(kwargs['tick']['last']),
                "lastsize": int(kwargs['tick']['size']),
                "bid": utils.to_decimal(kwargs['tick']['bid']),
                "ask": utils.to_decimal(kwargs['tick']['ask']),
                "bidsize": int(kwargs['tick']['bidsize']),
                "asksize": int(kwargs['tick']['asksize']),
                # "wap":          kwargs['tick']['wap'],
            }

        # for instruments that DOESN'T receive RTVOLUME events (exclude options)
        elif symbol not in self.rtvolume and \
                self.connection.contracts[tickerId].sec_type not in ("OPT", "FOP"):

            tick = self.connection.marketData[tickerId]
            # print(tick)

            if not tick.empty and tick['last'].values[-1] > 0 < tick['lastsize'].values[-1]:
                data = {
                    # available data from ib
                    "symbol": symbol,
                    "tickerId": str(tickerId),
                    # ES_F, ...
                    "symbol_group": utils.gen_symbol_group(symbol),
                    "asset_class": utils.gen_asset_class(symbol),
                    "timestamp": tick.index.values[-1],
                    "datetime": tick.index.values[-1],
                    "last": utils.to_decimal(tick['last'].values[-1]),
                    "lastsize": int(tick['lastsize'].values[-1]),
                    "buy": utils.to_decimal(tick['buy'].values[-1]),
                    "sell": utils.to_decimal(tick['sell'].values[-1]),
                    "buysize": int(tick['buysize'].values[-1]),
                    "sellsize": int(tick['sellsize'].values[-1])
                    # "wap":          kwargs['tick']['wap'],
                }

        # proceed if data exists
        if data is not None:
            # cache last tick
            if symbol in self.cash_ticks.keys():
                if data == self.cash_ticks[symbol]:
                    return

            self.cash_ticks[symbol] = data

            # add options fields
            # data = utils.force_options_columns(data)

            # print('.', end="", flush=True)
            self.on_tick_received(data)

    # -------------------------------------------
    @asynctools.multitasking.task
    def on_quote_received(self, tickerId):
        try:
            symbol = self.connection.tickerSymbol(tickerId)

            if self.connection.contracts[tickerId].sec_type in ("OPT", "FOP"):
                quote = self.connection.optionsData[tickerId].to_dict(orient='records')[
                    0]
                quote['type'] = self.connection.contracts[tickerId].right
                quote['strike'] = utils.to_decimal(
                    self.connection.contracts[tickerId].strike)
                quote["symbol_group"] = self.connection.contracts[tickerId].symbol + \
                                        '_' + self.connection.contracts[tickerId].sec_type
                quote = utils.mark_options_values(quote)
            else:
                quote = self.connection.marketQuoteData[tickerId].to_dict(orient='records')[0]
                quote["symbol_group"] = utils.gen_symbol_group(symbol)

            quote["symbol"] = symbol
            quote["tickerId"] = str(tickerId)
            quote["asset_class"] = utils.gen_asset_class(symbol)
            quote['bid'] = utils.to_decimal(quote['bid'])
            quote['ask'] = utils.to_decimal(quote['ask'])
            # quote['last'] = utils.to_decimal(quote['last'])
            quote["kind"] = "QUOTE"

            # cash markets do not get RTVOLUME (handleTickString)
            if quote["asset_class"] == "CSH":
                quote['last'] = round(
                    float((quote['bid'] + quote['ask']) / 2), 5)
                quote['timestamp'] = datetime.utcnow(
                ).strftime("%Y-%m-%d %H:%M:%S.%f")
                quote['datetime'] = quote['timestamp']
                # create synthetic tick
                if symbol in self.cash_ticks.keys() and quote['last'] != self.cash_ticks[symbol]:
                    self.on_tick_received(quote)
                else:
                    self.broadcast(quote, "QUOTE")

                self.cash_ticks[symbol] = quote['last']
            else:
                quoteStore = Tick(**quote)
                # print(quoteStore.to_json())
                self.log2db(quoteStore)
                self.broadcast(quote, "QUOTE")

        except Exception as e:
            pass

    # -------------------------------------------
    @asynctools.multitasking.task
    def on_option_computation_received(self, tickerId):
        # try:
        symbol = self.connection.tickerSymbol(tickerId)

        tick = self.connection.optionsData[tickerId].to_dict(orient='records')[0]

        # must have values!
        for key in ('bid', 'ask', 'last', 'bidsize', 'asksize', 'lastsize',
                    'volume', 'delta', 'gamma', 'vega', 'theta'):
            if tick[key] == 0:
                return

        tick['type'] = self.connection.contracts[tickerId].right
        tick['strike'] = utils.to_decimal(
            self.connection.contracts[tickerId].strike)
        tick["symbol_group"] = self.connection.contracts[tickerId].symbol + \
                               '_' + self.connection.contracts[tickerId].sec_type
        tick['volume'] = int(tick['volume'])
        tick['bid'] = utils.to_decimal(tick['bid'])
        tick['bidsize'] = int(tick['bidsize'])
        tick['ask'] = utils.to_decimal(tick['ask'])
        tick['asksize'] = int(tick['asksize'])
        tick['last'] = utils.to_decimal(tick['last'])
        tick['lastsize'] = int(tick['lastsize'])

        tick['price'] = utils.to_decimal(tick['price'], 2)
        tick['underlying'] = utils.to_decimal(tick['underlying'])
        tick['dividend'] = utils.to_decimal(tick['dividend'])
        tick['volume'] = int(tick['volume'])
        tick['iv'] = utils.to_decimal(tick['iv'])
        tick['oi'] = int(tick['oi'])
        tick['delta'] = utils.to_decimal(tick['delta'])
        tick['gamma'] = utils.to_decimal(tick['gamma'])
        tick['vega'] = utils.to_decimal(tick['vega'])
        tick['theta'] = utils.to_decimal(tick['theta'])

        tick["symbol"] = symbol
        tick["symbol_group"] = utils.gen_symbol_group(symbol)
        tick["asset_class"] = utils.gen_asset_class(symbol)

        tick = utils.mark_options_values(tick)

        # is this a really new tick?
        prev_last = 0
        prev_lastsize = 0
        if symbol in self.cash_ticks.keys():
            prev_last = self.cash_ticks[symbol]['last']
            prev_lastsize = self.cash_ticks[symbol]['lastsize']
            if tick == self.cash_ticks[symbol]:
                return

        self.cash_ticks[symbol] = dict(tick)

        # assign timestamp
        tick['timestamp'] = self.connection.optionsData[tickerId].index[0]
        if tick['timestamp'] == 0:
            tick['timestamp'] = datetime.utcnow().strftime(
                COMMON_TYPES['DATE_TIME_FORMAT_LONG_MILLISECS'])

        # treat as tick if last/volume changed
        if tick['last'] != prev_last or tick['lastsize'] != prev_lastsize:
            tick["kind"] = "TICK"
            self.on_tick_received(tick)

        # otherwise treat as quote
        else:
            tick["kind"] = "QUOTE"
            self.broadcast(tick, "QUOTE")

        # except Exception as e:
        # pass

    # -------------------------------------------
    @asynctools.multitasking.task
    def on_orderbook_received(self, tickerId):
        orderbook = self.connection.marketDepthData[tickerId].dropna(
            subset=['bid', 'ask']).fillna(0).to_dict(orient='list')

        # add symbol data to list
        symbol = self.connection.tickerSymbol(tickerId)
        orderbook['symbol'] = symbol
        orderbook["symbol_group"] = gen_symbol_group(symbol)
        orderbook["asset_class"] = gen_asset_class(symbol)
        orderbook["kind"] = "ORDERBOOK"

        quoteStore = Tick(tickerId=str(tickerId), symbol=symbol, kind="QUOTE", bid=float(orderbook['bid'][0]),
                          bidsize=int(orderbook['bidsize'][0]), ask=float(orderbook['ask'][0]),
                          asksize=int(orderbook['asksize'][0]))
        # print(quoteStore.to_json())
        self.log2db(quoteStore)
        # broadcast
        self.broadcast(orderbook, "ORDERBOOK")

    # -------------------------------------------
    @asynctools.multitasking.task
    def on_tick_received(self, tick):
        # data
        symbol = tick['symbol']
        timestamp = datetime.strptime(tick['timestamp'],
                                      COMMON_TYPES["DATE_TIME_FORMAT_LONG_MILLISECS"])

        # do not act on first tick (timezone is incorrect)
        if self.first_tick:
            self.first_tick = False
            return

        try:
            timestamp = parse_date(timestamp)
        except Exception as e:
            pass

        # placeholders
        if symbol not in self._raw_bars:
            self._raw_bars[symbol] = self._raw_bars['~']

        if symbol not in self._bars:
            self._bars[symbol] = self._bars['~']

        # send tick to message self.broadcast
        tick["kind"] = "TICK"
        self.broadcast(tick, "TICK")
        tickStore = Tick(**tick)
        self.log2db(tickStore)

        # add tick to raw self._bars
        tick_data = pd.DataFrame(index=['timestamp'],
                                 data={'timestamp': timestamp,
                                       'last': tick['last'],
                                       'volume': tick['lastsize']})
        tick_data.set_index(['timestamp'], inplace=True)
        _raw_bars = self._raw_bars[symbol].copy()
        _raw_bars = _raw_bars.append(tick_data)

        # add utils.resampled raw to self._bars
        ohlc = _raw_bars['last'].resample('1T').ohlc()
        vol = _raw_bars['volume'].resample('1T').sum()

        opened_bar = ohlc
        opened_bar['volume'] = vol

        # add bar to self._bars object
        previous_bar_count = len(self._bars[symbol])
        self._bars[symbol] = self._bars[symbol].append(opened_bar)
        self._bars[symbol] = self._bars[symbol].groupby(
            self._bars[symbol].index).last()

        if len(self._bars[symbol].index) > previous_bar_count:
            self.log_blotter.debug(f"__Bars__ {self._bars[symbol].to_dict(orient='records')[0]} \
            +{datetime.fromtimestamp(time.time())}")
            bar = self._bars[symbol].to_dict(orient='records')[0]
            bar["symbol"] = symbol
            bar["symbol_group"] = tick['symbol_group']
            bar["asset_class"] = tick['asset_class']
            bar["timestamp"] = self._bars[symbol].index[0].strftime(
                COMMON_TYPES["DATE_TIME_FORMAT_LONG"])

            bar["kind"] = "BAR"
            self.broadcast(bar, "BAR")
            barStore = OHLC(**bar, tickerId=tick['tickerId'])
            self.log2db(barStore)

            self._bars[symbol] = self._bars[symbol][-1:]
            _raw_bars.drop(_raw_bars.index[:], inplace=True)
        self._raw_bars[symbol] = _raw_bars

    # -------------------------------------------
    def broadcast(self, data, kind):
        def int64_handler(o):
            if isinstance(o, np_int64):
                try:
                    return pd.to_datetime(o, unit='ms').strftime(
                        COMMON_TYPES["DATE_TIME_FORMAT_LONG"])
                except Exception as e:
                    return int(o)
            raise TypeError

        string2send = "%s %s" % (
            self.args["zmqtopic"], json.dumps(data, default=int64_handler))

        # print(kind, string2send)
        try:
            self.socket.send_string(string2send)
        except Exception as e:
            pass

    # -------------------------------------------
    def log2db(self, data):
        try:
            if self.args['dbskip']:
                return
            if isinstance(data, Tick):
                return
            data.save()
        except Exception as e:
            self.log_blotter.error("Error inserting data into db %s", e)

    # -------------------------------------------
    def run(self):
        """Starts the blotter

        Connects to the TWS/GW, processes and logs market data,
        and broadcast it over TCP via ZeroMQ (which algo subscribe to)
        """

        self._check_unique_blotter()

        # connect to mysql
        self.db_connect()

        self.context = zmq.Context(zmq.REP)
        self.socket = self.context.socket(zmq.PUB)
        self.socket.bind("tcp://*:" + str(self.args['zmqport']))

        db_modified = 0
        contracts = []
        prev_contracts = []
        first_run = True

        self.log_blotter.info("Connecting to Broker...")

        self.connection = Webull()
        self.connection.callbacks = self.callbacks
        self.connection.connect(stream=True)

        self.log_blotter.info("Connection established...")

        try:
            while True:

                if not os.path.exists(self.args['symbols']):
                    pd.DataFrame(columns=['symbol', 'sec_type', 'exchange',
                                          'currency', 'expiry', 'strike', 'opt_type']
                                 ).to_csv(self.args['symbols'], header=True, index=False)
                    utils.chmod(self.args['symbols'])
                else:
                    time.sleep(0.1)

                    # read db properties
                    db_data = os.stat(self.args['symbols'])
                    db_size = db_data.st_size
                    db_last_modified = db_data.st_mtime

                    # empty file
                    if db_size == 0:
                        if prev_contracts:
                            self.log_blotter.info('Cancel market data...')
                            self.connection.cancelMarketDataSubscription()
                            time.sleep(0.1)
                            prev_contracts = []
                        continue

                    # modified?
                    if not first_run and db_last_modified == db_modified:
                        continue

                    # continue...
                    db_modified = db_last_modified

                    # read contructs db
                    df = pd.read_csv(self.args['symbols'], header=0)
                    if df.empty:
                        continue

                    # removed expired
                    df = df[(
                                    (df['expiry'] < 1000000) & (
                                    df['expiry'] >= int(datetime.now().strftime('%Y%m')))) | (
                                    (df['expiry'] >= 1000000) & (
                                    df['expiry'] >= int(datetime.now().strftime('%Y%m%d')))) |
                            np_isnan(df['expiry'])
                            ]

                    # fix expiry formatting (no floats)
                    df['expiry'] = df['expiry'].fillna(
                        0).astype(int).astype(str)
                    df.loc[df['expiry'] == "0", 'expiry'] = ""
                    df = df[df['sec_type'] != 'BAG']

                    df.fillna("", inplace=True)
                    df.to_csv(self.args['symbols'], header=True, index=False)
                    utils.chmod(self.args['symbols'])

                    # ignore comment
                    df = df[~df['symbol'].str.contains("#")]
                    contracts = [tuple(x) for x in df.values]

                    if first_run:
                        first_run = False

                    else:
                        if contracts != prev_contracts:
                            # cancel market data for removed contracts
                            for contract in prev_contracts:
                                if contract not in contracts:
                                    self.connection.cancelMarketDataSubscription(
                                        self.connection.createContract(contract))
                                    if self.args['orderbook']:
                                        self.connection.cancelMarketDepth(
                                            self.connection.createContract(contract))
                                    time.sleep(0.1)
                                    contract_string = self.connection.contractString(
                                        contract).split('_')[0]
                                    self.log_blotter.info(
                                        'Contract Removed [%s]', contract_string)

                    # request market data
                    for contract in contracts:
                        if contract not in prev_contracts:
                            self.connection.subscribeMarketData(
                                self.connection.createContract(contract))
                            if self.args['orderbook']:
                                self.connection.subscribeMarketDepth(
                                    self.connection.createContract(contract))
                            # time.sleep(0.1)
                            contract_string = self.connection.contractString(
                                contract).split('_')[0]
                            self.log_blotter.info(
                                'Contract Added [%s]', contract_string)

                    # update latest contracts
                    if prev_contracts != contracts:
                        if not self.connection.started:
                            self.connection.stream()

                    prev_contracts = contracts
                time.sleep(10)

        except (KeyboardInterrupt, SystemExit):
            self.quitting = True  # don't display connection errors on ctrl+c
            self.log_blotter.error(
                "\n\n>>> Interrupted with Ctrl-c...\n(waiting for running tasks to be completed)\n")
            # asynctools.multitasking.killall() # stop now
            asynctools.multitasking.wait_for_tasks()  # wait for threads to complete
            sys.exit(1)

    # -------------------------------------------
    # CLIENT / STATIC
    # -------------------------------------------
    def _fix_history_sequence(self, df, table):
        """ fix out-of-sequence ticks/bars """

        # remove "Unnamed: x" columns
        cols = df.columns[df.columns.str.startswith('Unnamed:')].tolist()
        df.drop(cols, axis=1, inplace=True)

        # remove future dates
        df['datetime'] = pd.to_datetime(df['datetime'], utc=True)
        blacklist = df[df['datetime'] > pd.to_datetime('now', utc=True)]
        df = df.loc[set(df.index) - set(blacklist)]  # .tail()

        # loop through data, symbol by symbol
        dfs = []
        bad_ids = [blacklist['_id'].values.tolist()]

        for symbol_id in list(df['symbol'].unique()):

            data = df[df['symbol'] == symbol_id].copy()

            # sort by id
            data.sort_values('datetime', axis=0, ascending=True, inplace=False)

            # convert index to column
            # data.loc[:, "ix"] = data.index
            # data.reset_index(inplace=True)

            # find out of sequence ticks/bars
            malformed = data.shift(1)[(data['resolution'] != data['resolution'].shift(1)) & (
                    data['datetime'] < data['datetime'].shift(1))]

            # cleanup rows
            if malformed.empty:
                # if all rows are in sequence, just remove last row
                dfs.append(data)
            else:
                # remove out of sequence rows + last row from data
                index = [
                    x for x in data.index.values if x not in malformed['datetime'].values]
                dfs.append(data.loc[index])

                # add to bad id list (to remove from db)
                bad_ids.append(list(malformed['_id'].values))

        # combine all lists
        data = pd.concat(dfs, sort=True)

        # flatten bad ids
        bad_ids = sum(bad_ids, [])

        # remove bad ids from db
        if bad_ids:
            bad_ids = list(map(str, map(int, bad_ids)))
            self.log_blotter.warning("Bad Ids found", bad_ids)

        # return
        return data.drop(['_id'], axis=1)

    # -------------------------------------------
    def history(self, symbols, start, end=None, resolution="1T", tz="UTC", continuous=True):
        # load runtime/default data
        if isinstance(symbols, str):
            symbols = symbols.split(',')

        # work with symbol groups
        # symbols = list(map(utils.gen_symbol_group, symbols))
        symbol_groups = list(map(utils.gen_symbol_group, symbols))
        # print(symbols)

        # convert datetime to string for MySQL
        try:
            start = start.strftime(
                COMMON_TYPES["DATE_TIME_FORMAT_LONG_MILLISECS"])
        except Exception as e:
            pass

        if end is not None:
            try:
                end = end.strftime(
                    COMMON_TYPES["DATE_TIME_FORMAT_LONG_MILLISECS"])
            except Exception as e:
                pass

        # connect to mysql
        self.db_connect()

        # --- build query
        table = 'ticks' if resolution[-1] in ("K", "V", "S") else 'bars'

        query = {
            'resolution': resolution
        }
        if symbols[0].strip() != "*":
            query['symbol__in'] = symbols

        from_to_query = (Q(datetime__gte=start) & Q(datetime__lte=end)) if end is not None else (Q(datetime__gte=start))
        data = OHLC.objects(from_to_query, **query).order_by('datetime')  # pd.read_sql(query, self.db_connection)  # .dropna()
        df = pd.DataFrame.from_records([json.loads(d.to_json()) for d in data])
        # del df['_id']
        if df.empty:
            return df
        df['_id'] = df['_id'].apply(lambda _id: _id['$oid'])
        df['datetime'] = df['datetime'].apply(lambda date: datetime.utcfromtimestamp(int(date['$date'] / 1000)))

        # print(df)
        # clearup records that are out of sequence
        data = self._fix_history_sequence(df, table)
        # setup dataframe
        data = prepare_history(data=data, resolution=Timeframes.timeframe_to_resolution(resolution), tz=tz,
                               continuous=continuous)

        # del query['resolution']
        # tick_data = Tick.objects(from_to_query, **query).order_by('datetime')
        # tick_df = pd.DataFrame.from_records([json.loads(d.to_json()) for d in tick_data])
        # tick_df['datetime'] = tick_df['datetime'].apply(lambda date: datetime.utcfromtimestamp(int(date['$date'] / 1000)))
        # tick_df = prepare_history(data=tick_df, tz=tz)
        # data['kind'] = 'BAR'
        # data = data.append(tick_df)
        # data.sort_index(inplace=True)
        # del data['_id']
        return data

    # -------------------------------------------
    def stream(self, symbols, tick_handler=None, bar_handler=None,
               quote_handler=None, book_handler=None, tz="UTC"):
        # load runtime/default data
        if isinstance(symbols, str):
            symbols = symbols.split(',')
        symbols = list(map(str.strip, symbols))

        # connect to zeromq self.socket
        self.context = zmq.Context()
        sock = self.context.socket(zmq.SUB)
        sock.setsockopt_string(zmq.SUBSCRIBE, "")
        uri = 'tcp://127.0.0.1:' + str(self.args['zmqport'])
        if os.getenv("zmq_connection_string"):
            uri = os.getenv("zmq_connection_string")
        sock.connect(uri)

        try:
            while True:
                message = sock.recv_string()

                if self.args["zmqtopic"] in message:
                    message = message.split(self.args["zmqtopic"])[1].strip()
                    data = json.loads(message)

                    if data['symbol'] not in symbols:
                        continue

                    # convert None to np.nan !!
                    data.update((k, np_nan)
                                for k, v in data.items() if v is None)

                    # quote
                    if data['kind'] == "ORDERBOOK":
                        if book_handler is not None:
                            try:
                                book_handler(data)
                            except Exception as e:
                                self.log_blotter.error(e)
                            continue
                    # quote
                    if data['kind'] == "QUOTE":
                        if quote_handler is not None:
                            try:
                                quote_handler(data)
                            except Exception as e:
                                self.log_blotter.error(e)
                            continue

                    try:
                        data["datetime"] = parse_date(data["timestamp"])
                    except Exception as e:
                        pass

                    df = pd.DataFrame(index=[0], data=data)
                    df.set_index('datetime', inplace=True)
                    df.index = pd.to_datetime(df.index, utc=True)
                    df.drop(["timestamp", "kind"], axis=1, inplace=True)

                    try:
                        df.index = df.index.tz_convert(tz)
                    except Exception as e:
                        df.index = df.index.tz_localize('UTC').tz_convert(tz)

                    # add options columns
                    df = force_options_columns(df)

                    if data['kind'] == "TICK":
                        if tick_handler is not None:
                            try:
                                tick_handler(df)
                            except Exception as e:
                                self.log_blotter.error(e)
                    elif data['kind'] == "BAR":
                        if bar_handler is not None:
                            try:
                                bar_handler(df)
                            except Exception as e:
                                self.log_blotter.error(e)

        except (KeyboardInterrupt, SystemExit):
            print(
                "\n\n>>> Interrupted with Ctrl-c...\n(waiting for running tasks to be completed)\n")
            print(".\n.\n.\n")
            # asynctools.multitasking.killall() # stop now
            asynctools.multitasking.wait_for_tasks()  # wait for threads to complete
            sys.exit(1)

    # -------------------------------------------
    @staticmethod
    def drip(data, handler):
        try:
            for i in range(len(data)):
                handler(data.iloc[i:i + 1])
                # time.sleep(.1)

            asynctools.multitasking.wait_for_tasks()
            print("\n\n>>> Backtesting Completed.")

        except (KeyboardInterrupt, SystemExit):
            print(
                "\n\n>>> Interrupted with Ctrl-c...\n(waiting for running tasks to be completed)\n")
            print(".\n.\n.\n")
            # asynctools.multitasking.killall() # stop now
            asynctools.multitasking.wait_for_tasks()  # wait for threads to complete
            sys.exit(1)

    # ---------------------------------------
    def backfill(self, data, resolution, start, end=None, csv_path=None):
        """
        Backfills missing historical data

        :Optional:
            data : pd.DataFrame
                Minimum required bars for backfill attempt
            resolution : str
                Algo resolution
            start: datetime
                Backfill start date (YYYY-MM-DD [HH:MM:SS[.MS]).
            end: datetime
                Backfill end date (YYYY-MM-DD [HH:MM:SS[.MS]). Default is None
        :Returns:
            status : mixed
                False for "won't backfill" / True for "backfilling, please wait"
        """

        data.sort_index(inplace=True)

        # currenly only supporting minute-data
        if resolution[-1] in ("K", "V"):
            self.backfilled = True
            return None

        # missing history?
        start_date = parse_date(start)
        end_date = parse_date(end) if end else datetime.utcnow()

        if data.empty:
            first_date = datetime.utcnow()
            last_date = datetime.utcnow()
        else:
            first_date = datetime64_to_datetime(data.index.values[0])
            last_date = datetime64_to_datetime(data.index.values[-1])

        self.backfill_resolution = Timeframes.to_timeframe(resolution)

        interval = Timeframes.timeframe_to_minutes(self.backfill_resolution)

        wb_lookback = None
        if start_date < first_date:
            wb_lookback = wb_lookback_str(start_date, end_date, interval)
        elif end_date > last_date:
            wb_lookback = wb_lookback_str(last_date, end_date, interval)

        if not wb_lookback:
            self.backfilled = True
            return None

        self.log_blotter.info("Backfilling historical data from IB...")

        # request parameters
        params = {
            "lookback": ceil(wb_lookback),
            "resolution": interval,
            "end_datetime": int(end_date.timestamp()),
            "csv_path": csv_path
        }

        # if connection is active - request data
        try:
            self.connection.requestHistoricalData(**params)
        except Exception as e:
            self.log_blotter.error(e)

        # wait for backfill to complete
        if not self.backfilled:
            raise Exception("Backfill interrupted")

        # otherwise, pass the parameters to the caller
        return True

    # -------------------------------------------
    def register(self, instruments):

        if isinstance(instruments, dict):
            instruments = list(instruments.values())

        if not isinstance(instruments, list):
            return

        try:
            db = pd.read_csv(self.args['symbols'], header=0).fillna("")

            instruments = pd.DataFrame(instruments)
            instruments.columns = db.columns
            # instruments['expiry'] = instruments['expiry'].astype(int).astype(str)

            db = db.append(instruments)
            # db['expiry'] = db['expiry'].astype(int)
            db = db.drop_duplicates(keep="first")

            db.to_csv(self.args['symbols'], header=True, index=False)
            chmod(self.args['symbols'])
        except Exception as e:
            self.log_blotter.error("Skipping symbols file since it couldn't be found in the system", e)

    # -------------------------------------------

    def db_connect(self):
        # skip db connection
        if self.args['dbskip']:
            return

        # already connected?
        if self.db_connection is not None:
            return

        # connect to mongo
        params = {
            'host': str(self.args['dbhost']) if str(self.args['dbhost']) is not None else 'localhost',
            'port': int(self.args['dbport']) if self.args['dbport'] is not None else 27017,
            'username': str(self.args['dbuser']) if str(self.args['dbuser']) is not None else None,
            'password': str(self.args['dbpass']) if str(self.args['dbpass']) is not None else None,
            'db': str(self.args['dbname']) if str(self.args['dbname']) else 'kinetick'
        }
        self.db_connection = mongo_connect(**params)

    # ===========================================
    # Utility functions --->
    # ===========================================

    # -------------------------------------------


# -------------------------------------------
def load_blotter_args(blotter_name=None, logger=None):
    """ Load running blotter's settings (used by clients)

    :Parameters:
        blotter_name : str
            Running Blotter's name (defaults to "auto-detect")
        logger : object
            Logger to be use (defaults to Blotter's)

    :Returns:
        args : dict
            Running Blotter's arguments
    """
    if logger is None:
        logger = create_logger(__name__, logging.WARNING)

    # find specific name
    if blotter_name is not None:  # and blotter_name != 'auto-detect':
        args_cache_file = tempfile.gettempdir() + "/" + blotter_name.lower() + ".kinetick"
        if not os.path.exists(args_cache_file):
            # logger.critical(
            #     "Cannot connect to running Blotter [%s]", blotter_name)
            # if os.isatty(0):
            #     sys.exit(0)
            return {}

    # no name provided - connect to last running
    else:
        blotter_files = sorted(
            glob.glob(tempfile.gettempdir() + "/*.kinetick"), key=os.path.getmtime)

        if not blotter_files:
            # logger.critical(
            #     "Cannot connect to running Blotter [%s]", blotter_name)
            # if os.isatty(0):
            #     sys.exit(0)
            return {}

        args_cache_file = blotter_files[-1]

    args = pickle.load(open(args_cache_file, "rb"))
    args['as_client'] = True

    return args


# -------------------------------------------
def prepare_history(data, resolution=None, tz="UTC", continuous=False):
    # setup dataframe
    data.set_index('datetime', inplace=True)
    data.index = pd.to_datetime(data.index, utc=True)
    # data['expiry'] = pd.to_datetime(data['expiry'], utc=True)

    # data['symbol'] = data['symbol'].str.replace("_STK", "")

    # force options columns
    # data = utils.force_options_columns(data)

    # construct continuous contracts for futures
    if continuous and resolution[-1] not in ("K", "V", "S"):
        all_dfs = [data[~data['symbol'].str.contains('FUT')]]

        # generate dict of df per future
        futures_symbol_groups = list(
            data[data['symbol'].str.contains('FUT')]['symbol'].unique())
        for key in futures_symbol_groups:
            future_group = data[data['symbol'] == key]
            continuous = create_continuous_contract(
                future_group, resolution)
            all_dfs.append(continuous)

        # make one df again
        # data = pd.concat(all_dfs, sort=True)
        data['datetime'] = data.index
        data.groupby([data.index, 'symbol'], as_index=False
                     ).last().set_index('datetime').dropna()

    if resolution is not None:
        data["symbol_group"] = data['symbol']
        data["asset_class"] = data['symbol']
        data = force_options_columns(data)
        data = resample(data, resolution, tz, sync_last_timestamp=False)
    return data


# -------------------------------------------
if __name__ == "__main__":
    blotter = Blotter()
    blotter.run()



import json
import time
from urllib.parse import urljoin

import dateutil.parser
import requests
import logging
import os

from kinetick.enums import SecurityType, PositionType

logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
                    level=os.getenv('LOGLEVEL') or logging.INFO)

logger = logging.getLogger(__name__)


class Zerodha():
    # Constants
    # Products
    PRODUCT_MIS = "MIS"
    PRODUCT_CNC = "CNC"
    PRODUCT_NRML = "NRML"
    PRODUCT_CO = "CO"
    PRODUCT_BO = "BO"

    # Order types
    ORDER_TYPE_MARKET = "MARKET"
    ORDER_TYPE_LIMIT = "LIMIT"
    ORDER_TYPE_SLM = "SL-M"
    ORDER_TYPE_SL = "SL"

    # Varities
    VARIETY_REGULAR = "regular"
    VARIETY_BO = "bo"
    VARIETY_CO = "co"
    VARIETY_AMO = "amo"

    # Transaction type
    TRANSACTION_TYPE_BUY = "BUY"
    TRANSACTION_TYPE_SELL = "SELL"

    # Validity
    VALIDITY_DAY = "DAY"
    VALIDITY_IOC = "IOC"

    # Exchanges
    EXCHANGE_NSE = "NSE"
    EXCHANGE_BSE = "BSE"
    EXCHANGE_NFO = "NFO"
    EXCHANGE_CDS = "CDS"
    EXCHANGE_BFO = "BFO"
    EXCHANGE_MCX = "MCX"

    # Margins segments
    MARGIN_EQUITY = "equity"
    MARGIN_COMMODITY = "commodity"

    # Status constants
    STATUS_COMPLETE = "COMPLETE"
    STATUS_REJECTED = "REJECTED"
    STATUS_CANCELLED = "CANCELLED"

    # GTT order type
    GTT_TYPE_OCO = "two-leg"
    GTT_TYPE_SINGLE = "single"

    # GTT order status
    GTT_STATUS_ACTIVE = "active"
    GTT_STATUS_TRIGGERED = "triggered"
    GTT_STATUS_DISABLED = "disabled"
    GTT_STATUS_EXPIRED = "expired"
    GTT_STATUS_CANCELLED = "cancelled"
    GTT_STATUS_REJECTED = "rejected"
    GTT_STATUS_DELETED = "deleted"

    _routes = {
        "login": "/api/login",
        "twofa": "/api/twofa",
        "api.token": "/session/token",
        "api.token.invalidate": "/session/token",
        "api.token.renew": "/session/refresh_token",
        "user.profile": "/user/profile",
        "user.margins": "/user/margins",
        "user.margins.segment": "/user/margins/{segment}",

        "orders": "oms/orders",
        "trades": "oms/trades",

        "order.info": "oms/orders/{order_id}",
        "order.place": "oms/orders/{variety}",
        "order.modify": "oms/orders/{variety}/{order_id}",
        "order.cancel": "oms/orders/{variety}/{order_id}",
        "order.trades": "oms/orders/{order_id}/trades",

        "portfolio.positions": "oms/portfolio/positions",
        "portfolio.holdings": "oms/portfolio/holdings",
        "portfolio.positions.convert": "oms/portfolio/positions",

        # MF api endpoints
        "mf.orders": "/mf/orders",
        "mf.order.info": "/mf/orders/{order_id}",
        "mf.order.place": "/mf/orders",
        "mf.order.cancel": "/mf/orders/{order_id}",

        "mf.sips": "/mf/sips",
        "mf.sip.info": "/mf/sips/{sip_id}",
        "mf.sip.place": "/mf/sips",
        "mf.sip.modify": "/mf/sips/{sip_id}",
        "mf.sip.cancel": "/mf/sips/{sip_id}",

        "mf.holdings": "/mf/holdings",
        "mf.instruments": "/mf/instruments",

        "market.instruments.all": "/instruments",
        "market.instruments": "/instruments/{exchange}",
        "market.margins": "/margins/{segment}",
        "market.historical": "/instruments/historical/{instrument_token}/{interval}",
        "market.trigger_range": "/instruments/trigger_range/{transaction_type}",

        "market.quote": "/quote",
        "market.quote.ohlc": "/quote/ohlc",
        "market.quote.ltp": "/quote/ltp",

        # GTT endpoints
        "gtt": "/gtt/triggers",
        "gtt.place": "/gtt/triggers",
        "gtt.info": "/gtt/triggers/{trigger_id}",
        "gtt.modify": "/gtt/triggers/{trigger_id}",
        "gtt.delete": "/gtt/triggers/{trigger_id}"
    }

    def __init__(self, user_id, password, pin, debug=False):
        self.user_id = user_id,
        self.password = password,
        self.pin = pin
        self.apiKey = 'kitefront'
        self.base_url = "https://kite.zerodha.com"

        self._session = requests.Session()
        self._session.headers.update({'X-Kite-Version': '2.4.0'})
        self.debug = debug
        self.timeout = 60
        self.maxretry = 3

        self._session_expiry_hook = self.default_session_expiry_hook

        # ==== set default values. =====
        self._account = {}
        self.orders = {}  # TODO initialize with pending orders
        self.symbol_orders = {}
        # =====

    def login(self):
        res = self._post("login", {'user_id': self.user_id, 'password': self.password})
        time.sleep(1)
        res = self._session.post(self.base_url + self._routes["twofa"],
                                 {'user_id': res['user_id'], 'request_id': res['request_id'],
                                  'twofa_value': self.pin})
        data = res.json()
        if data['status'] == 'success':
            logger.info("logged into zerodha")
            self._session.headers.update({'Authorization': "enctoken " + res.cookies.get_dict()['enctoken']})
            return

    def default_session_expiry_hook(self, response, **kwargs):
        logger.info("Running session expiry hook")
        headers = kwargs['headers'] if 'headers' in kwargs else {}
        retryAttempts = headers["x-retry"] if "x-retry" in headers else 1
        if int(retryAttempts) <= self.maxretry:
            logger.info(f"Retrying request. Attempt: {retryAttempts}")
            self.login()
            headers["x-retry"] = str(int(retryAttempts) + 1)
            kwargs['headers'] = headers
            return self._request(**kwargs)
        logger.error("Maximum session retry attempts {} exceeded".format(self.maxretry))
        raise Exception(f"zerodha: maximum re-login attempts {self.maxretry} failed")

    def set_session_expiry_hook(self, method):
        """
        Set a callback hook for session (`TokenError` -- timeout, expiry etc.) errors.
        A callback method that handles session errors
        can be set here and when the client encounters
        a token error at any point, it'll be called.

        This callback, for instance, can log the user out of the UI,
        clear session cookies, or initiate a fresh login.
        """
        if not callable(method):
            raise TypeError("Invalid input type. Only functions are accepted.")

        self._session_expiry_hook = method

    def place_order(self, variety, tradingsymbol, transaction_type, quantity, product,
                    order_type, exchange='NSE', **kwargs):
        """
        :param variety:
        :param tradingsymbol:
        :param transaction_type:
        :param quantity:
        :param product:
        :param order_type:
        :param exchange:
        :param price:
        :param trigger_price:
        :param validity:
        :param disclosed_quantity:
        :param trigger_price:
        :param squareoff:
        :param stoploss:
        :param squareoff:
        :param trailing_stoploss:
        :param tag:
        :return:
        """
        params = {'variety': variety,
                  'tradingsymbol': tradingsymbol,
                  'transaction_type': transaction_type,
                  'quantity': quantity,
                  'product': product,
                  'order_type': order_type,
                  'exchange': exchange
                  }
        for param in kwargs:
            if param is not None:
                params[param] = kwargs[param]
        response = self._post("order.place", params)
        logger.info("Order Placed with parameters ", params)
        return response["order_id"]

    def modify_order(self,
                     variety,
                     order_id,
                     parent_order_id=None,
                     quantity=None,
                     price=None,
                     order_type=None,
                     trigger_price=None,
                     validity=None,
                     disclosed_quantity=None):
        """Modify an open order."""
        params = locals()
        del (params["self"])

        for k in list(params.keys()):
            if params[k] is None:
                del (params[k])

        return self._put("order.modify", params)["order_id"]

    def cancel_order(self, variety, order_id, parent_order_id=None):
        """Cancel an order."""
        return self._delete("order.cancel", {
            "order_id": order_id,
            "variety": variety,
            "parent_order_id": parent_order_id
        })["order_id"]

    def exit_order(self, order_id, variety=None, parent_order_id=None):
        """Exit order."""
        if variety is None:
            order = self.order_by_id(order_id)[-1]
            variety = order['variety']
            parent_order_id = order['parent_order_id']
        self.cancel_order(variety, order_id, parent_order_id=parent_order_id)

    def _format_response(self, data):
        """Parse and format responses."""

        if type(data) == list:
            _list = data
        elif type(data) == dict:
            _list = [data]

        for item in _list:
            # Convert date time string to datetime object
            for field in ["order_timestamp", "exchange_timestamp", "created", "last_instalment", "fill_timestamp",
                          "timestamp", "last_trade_time"]:
                if item.get(field) and len(item[field]) == 19:
                    item[field] = dateutil.parser.parse(item[field])

        return _list[0] if type(data) == dict else _list

    def orders(self):
        """Get list of orders."""
        return self._format_response(self._get("orders"))

    def order_by_id(self, order_id):
        """
        Get history of individual order.
        """
        return self._format_response(self._get("order.info", {"order_id": order_id}))

    def positions(self):
        """Retrieve the list of positions."""
        return self._get("portfolio.positions")

    def holdings(self):
        """Retrieve the list of equity holdings."""
        return self._get("portfolio.holdings")

    def _get(self, route, params=None):
        """Alias for sending a GET request."""
        return self._request(route, "GET", params)

    def _post(self, route, params=None):
        """Alias for sending a POST request."""
        return self._request(route, "POST", params)

    def _put(self, route, params=None):
        """Alias for sending a PUT request."""
        return self._request(route, "PUT", params)

    def _delete(self, route, params=None):
        """Alias for sending a DELETE request."""
        return self._request(route, "DELETE", params)

    def _request(self, route, method, parameters=None, headers=None):
        """Make an HTTP request."""
        if headers is None:
            headers = {}
        params = parameters.copy() if parameters else {}

        # Form a restful URL
        uri = self._routes[route].format(**params)
        url = urljoin(self.base_url, uri)

        if self.debug:
            logger.debug("Request: {method} {url} {params}".format(method=method, url=url, params=params))

        try:
            response = self._session.request(method,
                                             url,
                                             data=params if method in ["POST", "PUT"] else None,
                                             params=params if method in ["GET", "DELETE"] else None,
                                             headers=headers,
                                             # verify=not self.disable_ssl,
                                             allow_redirects=True,
                                             timeout=self.timeout)
        # Any requests lib related exceptions are raised here -
        # http://docs.python-requests.org/en/master/_modules/requests/exceptions/
        except Exception as e:
            raise e

        if self.debug:
            logger.debug("Response: {code} {content}".format(code=response.status_code, content=response.content))

        # Validate the content type.
        if "json" in response.headers["content-type"]:
            try:
                data = json.loads(response.content.decode("utf8"))
            except ValueError:
                raise Exception("Couldn't parse the JSON response received from the server: {content}".format(
                    content=response.content))

            # api error
            if data.get("error_type"):
                # Call session hook if its registered and TokenException is raised
                if self._session_expiry_hook and response.status_code == 403 and data["error_type"] == "TokenException":
                    return self._session_expiry_hook(
                        response, route=route, method=method, parameters=parameters, headers=headers)
                else:
                    # native Kite errors
                    # exp = getattr(ex, data["error_type"], ex.GeneralException)
                    raise Exception(data["message"])

            return data["data"]
        elif "csv" in response.headers["content-type"]:
            return response.content
        else:
            raise Exception("Unknown Content-Type ({content_type}) with response: ({content})".format(
                content_type=response.headers["content-type"],
                content=response.content))

    @property
    def account(self):
        return self._account

    def get_order_variety(self, sec_type, pos_type):
        if sec_type == SecurityType.STOCK and pos_type == PositionType.CO:
            return Zerodha.VARIETY_CO, Zerodha.PRODUCT_MIS
        if sec_type == SecurityType.STOCK and pos_type == PositionType.MIS:
            return Zerodha.VARIETY_REGULAR, Zerodha.PRODUCT_MIS
        if sec_type == SecurityType.OPTION or sec_type == SecurityType.FUTURE:
            if pos_type == PositionType.CO or pos_type == PositionType.MIS:
                return Zerodha.VARIETY_REGULAR, Zerodha.PRODUCT_MIS
            return Zerodha.VARIETY_REGULAR, Zerodha.PRODUCT_NRML
        return Zerodha.VARIETY_REGULAR, Zerodha.PRODUCT_CNC


if __name__ == "__main__":
    # kite.set_access_token()
    zerodha = Zerodha("", "", "", debug=True)
    zerodha.login()
    # print(zerodha.getOrders())
    order = zerodha.place_order(variety=Zerodha.VARIETY_CO, tradingsymbol='ACC',
                                transaction_type=Zerodha.TRANSACTION_TYPE_BUY, quantity=1,
                                product=Zerodha.PRODUCT_CO, order_type=Zerodha.ORDER_TYPE_LIMIT,
                                price=89.5)
    print(order)



class sides:
    BUY = 'buy'
    SELL = 'sell'


class trade_types:
    LONG = 'long'
    SHORT = 'short'


class order_statuses:
    ACTIVE = 'ACTIVE'
    CANCELED = 'CANCELED'
    EXECUTED = 'EXECUTED'
    QUEUED = 'QUEUED'


class Timeframes:
    MINUTE_T = 'T'
    MINUTE_1 = '1m'
    MINUTE_3 = '3m'
    MINUTE_5 = '5m'
    MINUTE_15 = '15m'
    MINUTE_30 = '30m'
    HOUR_1 = '1h'
    HOUR_2 = '2h'
    HOUR_3 = '3h'
    HOUR_4 = '4h'
    HOUR_6 = '6h'
    HOUR_8 = '8h'
    DAY_1 = '1D'

    @staticmethod
    def tuple():
        return (
            Timeframes.MINUTE_T,
            Timeframes.MINUTE_1,
            Timeframes.MINUTE_3,
            Timeframes.MINUTE_5,
            Timeframes.MINUTE_15,
            Timeframes.MINUTE_30,
            Timeframes.HOUR_1,
            Timeframes.HOUR_2,
            Timeframes.HOUR_3,
            Timeframes.HOUR_4,
            Timeframes.HOUR_6,
            Timeframes.HOUR_8,
            Timeframes.DAY_1
        )

    @staticmethod
    def timeframe_to_minutes(timeframe):
        dic = {
            Timeframes.MINUTE_T: 1,
            Timeframes.MINUTE_1: 1,
            Timeframes.MINUTE_3: 3,
            Timeframes.MINUTE_5: 5,
            Timeframes.MINUTE_15: 15,
            Timeframes.MINUTE_30: 30,
            Timeframes.HOUR_1: 60,
            Timeframes.HOUR_2: 60 * 2,
            Timeframes.HOUR_3: 60 * 3,
            Timeframes.HOUR_4: 60 * 4,
            Timeframes.HOUR_6: 60 * 6,
            Timeframes.HOUR_8: 60 * 8,
            Timeframes.DAY_1: 60 * 24,
        }
        try:
            return dic[timeframe]
        except KeyError:
            raise Exception(
                'Timeframe "{}" is invalid. Supported timeframes are 1m, 3m, 5m, 15m, 30m, 1h, 2h, 3h, 4h, 6h, 8h, 1D'.format(
                    timeframe))

    @staticmethod
    def to_timeframe(minutes):
        if minutes in Timeframes.tuple():
            return minutes
        minutes = minutes if isinstance(minutes, int) else int(minutes.replace('T', ''))
        dic = {
            1: Timeframes.MINUTE_1,
            3: Timeframes.MINUTE_3,
            5: Timeframes.MINUTE_5,
            15: Timeframes.MINUTE_15,
            30: Timeframes.MINUTE_30,
            60: Timeframes.HOUR_1,
            120: Timeframes.HOUR_2,
            180: Timeframes.HOUR_3,
            240: Timeframes.HOUR_4,
            300: Timeframes.HOUR_6,
            360: Timeframes.HOUR_8,
            1440: Timeframes.DAY_1
        }
        try:
            return dic[minutes]
        except KeyError:
            raise Exception(
                'Minute "{}" is invalid. Supported timeframes are 1m, 3m, 5m, 15m, 30m, 1h, 2h, 3h, 4h, 6h, 8h, 1D'.format(
                    minutes))

    @staticmethod
    def timeframe_to_resolution(timeframe):
        minutes = Timeframes.timeframe_to_minutes(timeframe)
        return f'{minutes}T'

    @staticmethod
    def max_timeframe(timeframes_list):
        if Timeframes.DAY_1 in timeframes_list:
            return Timeframes.DAY_1
        if Timeframes.HOUR_8 in timeframes_list:
            return Timeframes.HOUR_8
        if Timeframes.HOUR_6 in timeframes_list:
            return Timeframes.HOUR_6
        if Timeframes.HOUR_4 in timeframes_list:
            return Timeframes.HOUR_4
        if Timeframes.HOUR_3 in timeframes_list:
            return Timeframes.HOUR_3
        if Timeframes.HOUR_2 in timeframes_list:
            return Timeframes.HOUR_2
        if Timeframes.HOUR_1 in timeframes_list:
            return Timeframes.HOUR_1
        if Timeframes.MINUTE_30 in timeframes_list:
            return Timeframes.MINUTE_30
        if Timeframes.MINUTE_15 in timeframes_list:
            return Timeframes.MINUTE_15
        if Timeframes.MINUTE_5 in timeframes_list:
            return Timeframes.MINUTE_5
        if Timeframes.MINUTE_3 in timeframes_list:
            return Timeframes.MINUTE_3

        return Timeframes.MINUTE_1


class colors:
    GREEN = 'green'
    YELLOW = 'yellow'
    RED = 'red'
    MAGENTA = 'magenta'
    BLACK = 'black'


class order_roles:
    OPEN_POSITION = 'OPEN POSITION'
    CLOSE_POSITION = 'CLOSE POSITION'
    INCREASE_POSITION = 'INCREASE POSITION'
    REDUCE_POSITION = 'REDUCE POSITION'


class order_flags:
    OCO = 'OCO'
    POST_ONLY = 'PostOnly'
    CLOSE = 'Close'
    HIDDEN = 'Hidden'
    REDUCE_ONLY = 'ReduceOnly'


class order_types:
    MARKET = 'MARKET'
    LIMIT = 'LIMIT'
    STOP = 'STOP'
    FOK = 'FOK'
    STOP_LIMIT = 'STOP LIMIT'


class PositionType:
    MIS = "MIS"  # Intraday with margin
    CO = "CO"  # Cover Order
    CNC = "CNC"  # Cash N Carry
    NRML = "NRML"  # derivative regular variety

    @staticmethod
    def get_all():
        return [PositionType.MIS, PositionType.CO, PositionType.CNC]

    @staticmethod
    def is_overnight_position(pos: str) -> bool:
        return pos.upper() == PositionType.CNC


class SecurityType:
    OPTION = "OPT"
    FUTURE = "FUT"
    STOCK = "STK"


class exchanges:
    COINBASE = 'Coinbase'
    BINANCE = 'Binance'
    BINANCE_FUTURES = 'Binance Futures'
    SANDBOX = 'Sandbox'
    BSE = 'BSE'
    NSE = 'NSE'
    NFO = 'NFO'


COMMON_TYPES = {
    "MONTH_CODES": ['', 'F', 'G', 'H', 'J', 'K', 'M', 'N', 'Q', 'U', 'V', 'X', 'Z'],

    "PRICE_TICKS": {1: "bid", 2: "ask", 4: "last", 6: "high", 7: "low", 9: "close", 14: "open"},
    "SIZE_TICKS": {0: "bid", 3: "ask", 5: "last", 8: "volume"},
    "RTVOL_TICKS": {"instrument": "", "ticketId": 0, "price": 0, "size": 0, "time": 0, "volume": 0, "vwap": 0,
                    "single": 0},

    "DURATION_1_HR": "3600 S",
    "DURATION_1_MIN": "60 S",
    "DURATION_1_DAY": "1 D",

    "BAR_SIZE_1_SEC": "1 secs",
    "BAR_SIZE_1_MIN": "1 min",

    "RTH_ALL": 0,
    "RTH_ONLY_TRADING_HRS": 1,

    "DATEFORMAT_STRING": 1,
    "DATEFORMAT_UNIX_TS": 2,

    "MSG_CURRENT_TIME": "currentTime",
    "MSG_COMMISSION_REPORT": "commissionReport",
    "MSG_CONNECTION_CLOSED": "connectionClosed",

    "MSG_CONTRACT_DETAILS": "contractDetails",
    "MSG_CONTRACT_DETAILS_END": "contractDetailsEnd",
    "MSG_TICK_SNAPSHOT_END": "tickSnapshotEnd",

    "MSG_TYPE_HISTORICAL_DATA": "historicalData",
    "MSG_TYPE_ACCOUNT_UPDATES": "updateAccountValue",
    "MSG_TYPE_ACCOUNT_TIME_UPDATES": "updateAccountTime",
    "MSG_TYPE_PORTFOLIO_UPDATES": "updatePortfolio",
    "MSG_TYPE_POSITION": "position",
    "MSG_TYPE_MANAGED_ACCOUNTS": "managedAccounts",

    "MSG_TYPE_NEXT_ORDER_ID": "nextValidId",
    "MSG_TYPE_OPEN_ORDER": "openOrder",
    "MSG_TYPE_ORDER_STATUS": "orderStatus",
    "MSG_TYPE_OPEN_ORDER_END": "openOrderEnd",

    "MSG_TYPE_MKT_DEPTH": "updateMktDepth",
    "MSG_TYPE_MKT_DEPTH_L2": "updateMktDepthL2",

    "MSG_TYPE_TICK_PRICE": "tickPrice",
    "MSG_TYPE_TICK_STRING": "tickString",
    "MSG_TYPE_TICK_GENERIC": "tickGeneric",
    "MSG_TYPE_TICK_SIZE": "tickSize",
    "MSG_TYPE_TICK_OPTION": "tickOptionComputation",

    "DATE_FORMAT": "%Y%m%d",
    "DATE_TIME_FORMAT": "%Y%m%d %H:%M:%S",
    "DATE_TIME_FORMAT_LONG": "%Y-%m-%d %H:%M:%S",
    "DATE_TIME_FORMAT_LONG_MILLISECS": "%Y-%m-%d %H:%M:%S.%f",
    "DATE_TIME_FORMAT_HISTORY": "%Y%m%d %H:%M:%S",
    "DATE_FORMAT_HISTORY": "%Y-%m-%d",

    "GENERIC_TICKS_NONE": "",
    "GENERIC_TICKS_RTVOLUME": "233",

    "SNAPSHOT_NONE": False,
    "SNAPSHOT_TRUE": True,

    # MARKET ORDERS
    "ORDER_TYPE_MARKET": "MKT",
    "ORDER_TYPE_MOC": "MOC",  # MARKET ON CLOSE
    "ORDER_TYPE_MOO": "MOO",  # MARKET ON OPEN
    "ORDER_TYPE_MIT": "MIT",  # MARKET IF TOUCHED

    # LIMIT ORDERS
    "ORDER_TYPE_LIMIT": "LMT",
    "ORDER_TYPE_LOC": "LOC",  # LIMIT ON CLOSE
    "ORDER_TYPE_LOO": "LOO",  # LIMIT ON OPEN
    "ORDER_TYPE_LIT": "LIT",  # LIMIT IF TOUCHED

    # STOP ORDERS
    "ORDER_TYPE_STOP": "STP",
    "ORDER_TYPE_STOP_LIMIT": "STP LMT",
    "ORDER_TYPE_TRAIL_STOP": "TRAIL",
    "ORDER_TYPE_TRAIL_STOP_LIMIT": "TRAIL LIMIT",
    "ORDER_TYPE_TRAIL_STOP_LIT": "TRAIL LIT",  # LIMIT IF TOUCHED
    "ORDER_TYPE_TRAIL_STOP_MIT": "TRAIL MIT",  # MARKET IF TOUCHED

    # MINC ORDERS
    "ORDER_TYPE_ONE_CANCELS_ALL": "OCA",
    "ORDER_TYPE_RELATIVE": "REL",
    "ORDER_TYPE_PEGGED_TO_MIDPOINT": "PEG MID",

    # ORDER ACTIONS
    "ORDER_ACTION_SELL": "SELL",
    "ORDER_ACTION_BUY": "BUY",
    "ORDER_ACTION_SHORT": "SSHORT"
}



from mongoengine import StringField, DateTimeField, FloatField, IntField, DynamicDocument
import datetime

tick_types = ('TICK', 'QUOTE')


class Tick(DynamicDocument):
    tickerId = StringField(max_length=50, required=True)
    symbol = StringField(max_length=50, required=False)
    datetime = DateTimeField(required=True, default=datetime.datetime.utcnow)
    open = FloatField(required=False)
    high = FloatField(required=False)
    low = FloatField(required=False)
    close = FloatField(required=False)
    volume = IntField(required=False)
    vwap = FloatField()
    buy = FloatField()
    buysize = IntField()
    sell = FloatField()
    sellsize = IntField()
    last = FloatField()
    lastSize = IntField()
    ask = FloatField()
    asksize = IntField()
    bid = FloatField()
    bidSize = IntField()
    kind = StringField(default="TICK", choices=tick_types)
    meta = {
        'indexes': ['tickerId', 'datetime', 'kind']
    }



# Copyright 2021 vin8tech, vinay patil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import traceback
from abc import ABCMeta

from telegram import InlineKeyboardButton, InlineKeyboardMarkup
from telegram.ext import Updater, CommandHandler, CallbackQueryHandler

from kinetick.bot.dummy import DumbBot
from kinetick.utils.utils import rand_pass

logger = logging.getLogger(__name__)
logger.setLevel(os.getenv('LOGLEVEL') or logging.INFO)

__TOKEN__ = os.getenv("BOT_TOKEN")
__CHAT_ID__ = os.getenv("CHAT_ID")


class TelegramBot(DumbBot):
    _instance = None

    __metaclass__ = ABCMeta

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(TelegramBot, cls).__new__(cls)
            TelegramBot.init(cls._instance)
        return cls._instance

    @staticmethod
    def init(self):
        self._chat_ids = set()
        self._verified_chat_id = None  # only verified user can place orders.
        self._password = rand_pass(6)
        self._on_connected_listeners = set()
        self._callbacks_store = {}
        self._order_number = 0
        if __TOKEN__:
            self.bot = Updater(__TOKEN__, use_context=True)
            self.bot.dispatcher.add_handler(CommandHandler("start", self._start_cmd_handler))
            self.bot.dispatcher.add_handler(CommandHandler("stop", self._stop_cmd_handler))
            self.bot.dispatcher.add_handler(CommandHandler("login", self._login_cmd_handler))
            self.bot.dispatcher.add_handler(CallbackQueryHandler(self._button))

            logger.info("starting bot..")
            try:
                self.bot.start_polling()
                logger.debug("telegram bot started polling ...")
                logger.info("Use Login Pass to login into bot.")
                print(f"Login Pass: {self._password}")
            except Exception as e:
                logger.error("Failed to launch bot", e)
        if __CHAT_ID__ is not None:
            self._chat_ids.add(int(__CHAT_ID__))
            self._verified_chat_id = int(__CHAT_ID__)

    def start(self, *args, **kwargs):
        pass

    def stop(self, *args, **kwargs):
        if self.bot:
            self.bot.stop()

    def send_message(self, msg):
        for user in self._chat_ids:
            self.bot.bot.send_message(text=msg, chat_id=user, parse_mode='Markdown')

    def send_order(self, position, signal, callback=None, commands: tuple = (), **kwargs):
        self._order_number += 1
        key = signal + str(self._order_number)
        keyboard = [list(map(lambda cmd: InlineKeyboardButton(str(cmd), callback_data=key + f"${cmd}"), commands))]
        # cmd used in _button handler to pass the selected option to callback.

        reply_markup = InlineKeyboardMarkup(keyboard)

        tradingview_symbol = position.symbol
        tradingview_symbol = 'NIFTY' if tradingview_symbol == 'NSEI' else tradingview_symbol
        tradingview_symbol = 'BANKNIFTY' if tradingview_symbol == 'NSEBANK' else tradingview_symbol

        message = f'Symbol: [{position.symbol}](https://tradingview.com/chart/?symbol={tradingview_symbol}) \n ' \
                  f'Signal: {signal} \n Direction: *#{position.direction}* \n ' \
                  f'Strategy #{position.algo} \n' \
                  f'Entry Price:  {position.entry_price:.2f}  \n Quantity:  {position.quantity} \n ' \
                  f'Stoploss: {position.stop:.2f} \n Target: {position.target:.2f} \n ' \
                  f'Exit Price: {position.exit_price:.2f} \n'
        for user in self._chat_ids:
            self.bot.bot.send_message(text=message, chat_id=user, parse_mode='Markdown')
            if user == self._verified_chat_id:
                self.bot.bot.send_message(text=f'{position.symbol} {position.direction} confirm?',
                                          reply_markup=reply_markup, chat_id=user)
        if callback is not None:
            if callable(callback):
                self._callbacks_store[key] = callback

    def add_connected_listener(self, listener):
        self._on_connected_listeners.add(listener)

    def add_command_handler(self, command, handler, help_string=""):
        self.bot.dispatcher.add_handler(CommandHandler(command, handler))

    def _start_cmd_handler(self, update, context):
        update.message.reply_text('Hi {}, Use `/login <password>` command to start trading.'
                                  .format(update.message.from_user.first_name), parse_mode='Markdown')
        self._chat_ids.add(update.message.chat_id)
        for listener in self._on_connected_listeners:
            listener(update.message.chat_id)

    def _login_cmd_handler(self, update, ctx):
        text = update.message.text.split()[-1]
        if text == self._password:
            self._chat_ids.add(update.message.chat_id)
            self._verified_chat_id = update.message.chat_id
            msg = "Successfully logged in. Happy trading!"
            self.bot.bot.send_message(text=msg, chat_id=self._verified_chat_id)
        else:
            msg = "Invalid password!"
            self.bot.bot.send_message(text=msg, chat_id=update.message.chat_id)

    def _stop_cmd_handler(self, update, ctx):
        chat_id = update.message.chat_id
        if chat_id in self._chat_ids:
            self._chat_ids.remove(chat_id)

    def _button(self, update, context):
        query = update.callback_query
        query.answer()
        if self._verified_chat_id == update.effective_chat.id:
            data, *cmd = query.data.split("$")
            cmd = cmd[-1] if len(cmd) > 0 else data
            callback = self._callbacks_store[data] if data in self._callbacks_store else None

            if callback is not None:
                del self._callbacks_store[data]
                try:
                    callback(commands=(cmd, ))
                    query.edit_message_text(text="{} order request sent".format(cmd))
                except Exception as e:
                    traceback.print_exc()
                    logger.error('Error executing command %s', e)
                    query.edit_message_text(text="Error sending order request. Reason: {}".format(e))
        else:
            query.edit_message_text(text="Can not execute order.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# QTPyLib: Quantitative Trading Python Library
# https://github.com/ranaroussi/qtpylib
#
# Copyright 2016-2018 Ran Aroussi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import warnings
import sys
from datetime import datetime, timedelta

import numpy as np
import pandas as pd
from pandas.core.base import PandasObject

# =============================================
# check min, python version
from .supertrend import SuperTrend

# =============================================
warnings.simplefilter(action="ignore", category=RuntimeWarning)


# =============================================


def numpy_rolling_window(data, window):
    shape = data.shape[:-1] + (data.shape[-1] - window + 1, window)
    strides = data.strides + (data.strides[-1],)
    return np.lib.stride_tricks.as_strided(data, shape=shape, strides=strides)


def numpy_rolling_series(func):
    def func_wrapper(data, window, as_source=False):
        series = data.values if isinstance(data, pd.Series) else data

        new_series = np.empty(len(series)) * np.nan
        calculated = func(series, window)
        new_series[-len(calculated):] = calculated

        if as_source and isinstance(data, pd.Series):
            return pd.Series(index=data.index, data=new_series)

        return new_series

    return func_wrapper


@numpy_rolling_series
def numpy_rolling_mean(data, window, as_source=False):
    return np.mean(numpy_rolling_window(data, window), axis=-1)


@numpy_rolling_series
def numpy_rolling_std(data, window, as_source=False):
    return np.std(numpy_rolling_window(data, window), axis=-1, ddof=1)


# ---------------------------------------------


def session(df, start='17:00', end='16:00'):
    """ remove previous globex day from df """
    if df.empty:
        return df

    # get start/end/now as decimals
    int_start = list(map(int, start.split(':')))
    int_start = (int_start[0] + int_start[1] - 1 / 100) - 0.0001
    int_end = list(map(int, end.split(':')))
    int_end = int_end[0] + int_end[1] / 100
    int_now = (df[-1:].index.hour[0] + (df[:1].index.minute[0]) / 100)

    # same-dat session?
    is_same_day = int_end > int_start

    # set pointers
    curr = prev = df[-1:].index[0].strftime('%Y-%m-%d')

    # globex/forex session
    if not is_same_day:
        prev = (datetime.strptime(curr, '%Y-%m-%d') -
                timedelta(1)).strftime('%Y-%m-%d')

    # slice
    if int_now >= int_start:
        df = df[df.index >= curr + ' ' + start]
    else:
        df = df[df.index >= prev + ' ' + start]

    return df.copy()


# ---------------------------------------------


def heikinashi(bars):
    bars = bars.copy()
    bars['ha_close'] = (bars['open'] + bars['high'] +
                        bars['low'] + bars['close']) / 4

    # ha open
    bars.at[0, 'ha_open'] = (bars.at[0, 'open'] + bars.at[0, 'close']) / 2
    for i in range(1, len(bars)):
        bars.at[i, 'ha_open'] = (bars.at[i - 1, 'ha_open'] + bars.at[i - 1, 'ha_close']) / 2

    bars['ha_high'] = bars.loc[:, ['high', 'ha_open', 'ha_close']].max(axis=1)
    bars['ha_low'] = bars.loc[:, ['low', 'ha_open', 'ha_close']].min(axis=1)

    return pd.DataFrame(index=bars.index,
                        data={'open': bars['ha_open'],
                              'high': bars['ha_high'],
                              'low': bars['ha_low'],
                              'close': bars['ha_close']})


# ---------------------------------------------


def tdi(series, rsi_lookback=13, rsi_smooth_len=2,
        rsi_signal_len=7, bb_lookback=34, bb_std=1.6185):
    rsi_data = rsi(series, rsi_lookback)
    rsi_smooth = sma(rsi_data, rsi_smooth_len)
    rsi_signal = sma(rsi_data, rsi_signal_len)

    bb_series = bollinger_bands(rsi_data, bb_lookback, bb_std)

    return pd.DataFrame(index=series.index, data={
        "rsi": rsi_data,
        "rsi_signal": rsi_signal,
        "rsi_smooth": rsi_smooth,
        "rsi_bb_upper": bb_series['upper'],
        "rsi_bb_lower": bb_series['lower'],
        "rsi_bb_mid": bb_series['mid']
    })


# ---------------------------------------------


def awesome_oscillator(df, weighted=False, fast=5, slow=34):
    midprice = (df['high'] + df['low']) / 2

    if weighted:
        ao = (midprice.ewm(fast).mean() - midprice.ewm(slow).mean()).values
    else:
        ao = numpy_rolling_mean(midprice, fast) - \
             numpy_rolling_mean(midprice, slow)

    return pd.Series(index=df.index, data=ao)


# ---------------------------------------------

def nans(length=1):
    mtx = np.empty(length)
    mtx[:] = np.nan
    return mtx


# ---------------------------------------------

def typical_price(bars):
    res = (bars['high'] + bars['low'] + bars['close']) / 3.
    return pd.Series(index=bars.index, data=res)


# ---------------------------------------------

def mid_price(bars):
    res = (bars['high'] + bars['low']) / 2.
    return pd.Series(index=bars.index, data=res)


# ---------------------------------------------

def ibs(bars):
    """ Internal bar strength """
    res = np.round((bars['close'] - bars['low']) /
                   (bars['high'] - bars['low']), 2)
    return pd.Series(index=bars.index, data=res)


# ---------------------------------------------

def true_range(bars):
    return pd.DataFrame({
        "hl": bars['high'] - bars['low'],
        "hc": abs(bars['high'] - bars['close'].shift(1)),
        "lc": abs(bars['low'] - bars['close'].shift(1))
    }).max(axis=1)


# ---------------------------------------------

def atr(bars, window=14, exp=False):
    tr = true_range(bars)

    if exp:
        res = rolling_weighted_mean(tr, window)
    else:
        res = rolling_mean(tr, window)

    return pd.Series(res)


# ---------------------------------------------

def crossed(series1, series2, direction=None):
    if isinstance(series1, np.ndarray):
        series1 = pd.Series(series1)

    if isinstance(series2, (float, int, np.ndarray)):
        series2 = pd.Series(index=series1.index, data=series2)

    if direction is None or direction == "above":
        above = pd.Series((series1 > series2) & (
                series1.shift(1) <= series2.shift(1)))

    if direction is None or direction == "below":
        below = pd.Series((series1 < series2) & (
                series1.shift(1) >= series2.shift(1)))

    if direction is None:
        return above or below

    return above if direction == "above" else below


def crossed_above(series1, series2):
    return crossed(series1, series2, "above")


def crossed_below(series1, series2):
    return crossed(series1, series2, "below")


# ---------------------------------------------


def rolling_std(series, window=200, min_periods=None):
    min_periods = window if min_periods is None else min_periods
    if min_periods == window and len(series) > window:
        return numpy_rolling_std(series, window, True)
    else:
        try:
            return series.rolling(window=window, min_periods=min_periods).std()
        except Exception as e:
            return pd.Series(series).rolling(window=window, min_periods=min_periods).std()


# ---------------------------------------------


def rolling_mean(series, window=200, min_periods=None):
    min_periods = window if min_periods is None else min_periods
    if min_periods == window and len(series) > window:
        return numpy_rolling_mean(series, window, True)
    else:
        try:
            return series.rolling(window=window, min_periods=min_periods).mean()
        except Exception as e:
            return pd.Series(series).rolling(window=window, min_periods=min_periods).mean()


# ---------------------------------------------


def rolling_min(series, window=14, min_periods=None):
    min_periods = window if min_periods is None else min_periods
    try:
        return series.rolling(window=window, min_periods=min_periods).min()
    except Exception as e:
        return pd.Series(series).rolling(window=window, min_periods=min_periods).min()


# ---------------------------------------------

def rolling_max(series, window=14, min_periods=None):
    min_periods = window if min_periods is None else min_periods
    try:
        return series.rolling(window=window, min_periods=min_periods).min()
    except Exception as e:
        return pd.Series(series).rolling(window=window, min_periods=min_periods).min()


# ---------------------------------------------

def rolling_weighted_mean(series, window=200, min_periods=None):
    min_periods = window if min_periods is None else min_periods
    try:
        return series.ewm(span=window, min_periods=min_periods).mean()
    except Exception as e:
        return pd.ewma(series, span=window, min_periods=min_periods)


# ---------------------------------------------

def hull_moving_average(series, window=200, min_periods=None):
    min_periods = window if min_periods is None else min_periods
    ma = (2 * rolling_weighted_mean(series, window / 2, min_periods)) - \
         rolling_weighted_mean(series, window, min_periods)
    return rolling_weighted_mean(ma, np.sqrt(window), min_periods)


# ---------------------------------------------

def sma(series, window=200, min_periods=None):
    return rolling_mean(series, window=window, min_periods=min_periods)


# ---------------------------------------------

def wma(series, window=200, min_periods=None):
    return rolling_weighted_mean(series, window=window, min_periods=min_periods)


# ---------------------------------------------

def hma(series, window=200, min_periods=None):
    return hull_moving_average(series, window=window, min_periods=min_periods)


# ---------------------------------------------

def vwap(bars):
    """
    calculate vwap of entire time series
    (input can be pandas series or numpy array)
    bars are usually mid [ (h+l)/2 ] or typical [ (h+l+c)/3 ]
    """
    typical = ((bars['high'] + bars['low'] + bars['close']) / 3).values
    volume = bars['volume'].values

    return pd.Series(index=bars.index,
                     data=np.cumsum(volume * typical) / np.cumsum(volume))


# ---------------------------------------------

def rolling_vwap(bars, window=200, min_periods=None):
    """
    calculate vwap using moving window
    (input can be pandas series or numpy array)
    bars are usually mid [ (h+l)/2 ] or typical [ (h+l+c)/3 ]
    """
    min_periods = window if min_periods is None else min_periods

    typical = ((bars['high'] + bars['low'] + bars['close']) / 3)
    volume = bars['volume']

    left = (volume * typical).rolling(window=window,
                                      min_periods=min_periods).sum()
    right = volume.rolling(window=window, min_periods=min_periods).sum()

    return pd.Series(index=bars.index, data=(left / right)).replace([np.inf, -np.inf], float('NaN')).ffill()


# ---------------------------------------------

def rsi(series, window=14):
    """
    compute the n period relative strength indicator
    """

    # 100-(100/relative_strength)
    deltas = np.diff(series)
    seed = deltas[:window + 1]

    # default values
    ups = seed[seed > 0].sum() / window
    downs = -seed[seed < 0].sum() / window
    rsival = np.zeros_like(series)
    rsival[:window] = 100. - 100. / (1. + ups / downs)

    # period values
    for i in range(window, len(series)):
        delta = deltas[i - 1]
        if delta > 0:
            upval = delta
            downval = 0
        else:
            upval = 0
            downval = -delta

        ups = (ups * (window - 1) + upval) / window
        downs = (downs * (window - 1.) + downval) / window
        rsival[i] = 100. - 100. / (1. + ups / downs)

    # return rsival
    return pd.Series(index=series.index, data=rsival)


# ---------------------------------------------

def macd(series, fast=3, slow=10, smooth=16):
    """
    compute the MACD (Moving Average Convergence/Divergence)
    using a fast and slow exponential moving avg'
    return value is emaslow, emafast, macd which are len(x) arrays
    """
    macd_line = rolling_weighted_mean(series, window=fast) - \
                rolling_weighted_mean(series, window=slow)
    signal = rolling_weighted_mean(macd_line, window=smooth)
    histogram = macd_line - signal
    # return macd_line, signal, histogram
    return pd.DataFrame(index=series.index, data={
        'macd': macd_line.values,
        'signal': signal.values,
        'histogram': histogram.values
    })


# ---------------------------------------------

def bollinger_bands(series, window=20, stds=2):
    ma = rolling_mean(series, window=window, min_periods=1)
    std = rolling_std(series, window=window, min_periods=1)
    upper = ma + std * stds
    lower = ma - std * stds

    return pd.DataFrame(index=series.index, data={
        'upper': upper,
        'mid': ma,
        'lower': lower
    })


# ---------------------------------------------

def weighted_bollinger_bands(series, window=20, stds=2):
    ema = rolling_weighted_mean(series, window=window)
    std = rolling_std(series, window=window)
    upper = ema + std * stds
    lower = ema - std * stds

    return pd.DataFrame(index=series.index, data={
        'upper': upper.values,
        'mid': ema.values,
        'lower': lower.values
    })


# ---------------------------------------------

def returns(series):
    try:
        res = (series / series.shift(1) -
               1).replace([np.inf, -np.inf], float('NaN'))
    except Exception as e:
        res = nans(len(series))

    return pd.Series(index=series.index, data=res)


# ---------------------------------------------

def log_returns(series):
    try:
        res = np.log(series / series.shift(1)
                     ).replace([np.inf, -np.inf], float('NaN'))
    except Exception as e:
        res = nans(len(series))

    return pd.Series(index=series.index, data=res)


# ---------------------------------------------

def implied_volatility(series, window=252):
    try:
        logret = np.log(series / series.shift(1)
                        ).replace([np.inf, -np.inf], float('NaN'))
        res = numpy_rolling_std(logret, window) * np.sqrt(window)
    except Exception as e:
        res = nans(len(series))

    return pd.Series(index=series.index, data=res)


# ---------------------------------------------

def keltner_channel(bars, window=14, atrs=2):
    typical_mean = rolling_mean(typical_price(bars), window)
    atrval = atr(bars, window) * atrs

    upper = typical_mean + atrval
    lower = typical_mean - atrval

    return pd.DataFrame(index=bars.index, data={
        'upper': upper.values,
        'mid': typical_mean.values,
        'lower': lower.values
    })


# ---------------------------------------------

def roc(series, window=14):
    """
    compute rate of change
    """
    res = (series - series.shift(window)) / series.shift(window)
    return pd.Series(index=series.index, data=res)


# ---------------------------------------------

def cci(series, window=14):
    """
    compute commodity channel index
    """
    price = typical_price(series)
    typical_mean = rolling_mean(price, window)
    res = (price - typical_mean) / (.015 * np.std(typical_mean))
    return pd.Series(index=series.index, data=res)


# ---------------------------------------------

def stoch(df, window=14, d=3, k=3, fast=False):
    """
    compute the n period relative strength indicator
    http://excelta.blogspot.co.il/2013/09/stochastic-oscillator-technical.html
    """

    my_df = pd.DataFrame(index=df.index)

    my_df['rolling_max'] = df['high'].rolling(window).max()
    my_df['rolling_min'] = df['low'].rolling(window).min()

    my_df['fast_k'] = 100 * (df['close'] - my_df['rolling_min']) / (my_df['rolling_max'] - my_df['rolling_min'])
    my_df['fast_d'] = my_df['fast_k'].rolling(d).mean()

    if fast:
        return my_df.loc[:, ['fast_k', 'fast_d']]

    my_df['slow_k'] = my_df['fast_k'].rolling(k).mean()
    my_df['slow_d'] = my_df['slow_k'].rolling(d).mean()

    return my_df.loc[:, ['slow_k', 'slow_d']]


# ---------------------------------------------


def zlma(series, window=20, min_periods=None, kind="ema"):
    """
    John Ehlers' Zero lag (exponential) moving average
    https://en.wikipedia.org/wiki/Zero_lag_exponential_moving_average
    """
    min_periods = window if min_periods is None else min_periods

    lag = (window - 1) // 2
    series = 2 * series - series.shift(lag)
    if kind in ['ewm', 'ema']:
        return wma(series, lag, min_periods)
    elif kind == "hma":
        return hma(series, lag, min_periods)
    return sma(series, lag, min_periods)


def zlema(series, window, min_periods=None):
    return zlma(series, window, min_periods, kind="ema")


def zlsma(series, window, min_periods=None):
    return zlma(series, window, min_periods, kind="sma")


def zlhma(series, window, min_periods=None):
    return zlma(series, window, min_periods, kind="hma")


# ---------------------------------------------


def zscore(bars, window=20, stds=1, col='close'):
    """ get zscore of price """
    std = numpy_rolling_std(bars[col], window)
    mean = numpy_rolling_mean(bars[col], window)
    return (bars[col] - mean) / (std * stds)


# ---------------------------------------------


def pvt(bars):
    """ Price Volume Trend """
    trend = ((bars['close'] - bars['close'].shift(1)) /
             bars['close'].shift(1)) * bars['volume']
    return trend.cumsum()


def chopiness(bars, window=14):
    atrsum = true_range(bars).rolling(window).sum()
    highs = bars['high'].rolling(window).max()
    lows = bars['low'].rolling(window).min()
    return 100 * np.log10(atrsum / (highs - lows)) / np.log10(window)


def supertrend(bars, period=7, multiplier=3):
    return SuperTrend(bars.copy(), period=period, multiplier=multiplier)


# =============================================


PandasObject.session = session
PandasObject.atr = atr
PandasObject.bollinger_bands = bollinger_bands
PandasObject.cci = cci
PandasObject.crossed = crossed
PandasObject.crossed_above = crossed_above
PandasObject.crossed_below = crossed_below
PandasObject.heikinashi = heikinashi
PandasObject.hull_moving_average = hull_moving_average
PandasObject.ibs = ibs
PandasObject.implied_volatility = implied_volatility
PandasObject.keltner_channel = keltner_channel
PandasObject.log_returns = log_returns
PandasObject.macd = macd
PandasObject.returns = returns
PandasObject.roc = roc
PandasObject.rolling_max = rolling_max
PandasObject.rolling_min = rolling_min
PandasObject.rolling_mean = rolling_mean
PandasObject.rolling_std = rolling_std
PandasObject.rsi = rsi
PandasObject.stoch = stoch
PandasObject.zscore = zscore
PandasObject.pvt = pvt
PandasObject.chopiness = chopiness
PandasObject.tdi = tdi
PandasObject.true_range = true_range
PandasObject.mid_price = mid_price
PandasObject.typical_price = typical_price
PandasObject.vwap = vwap
PandasObject.rolling_vwap = rolling_vwap
PandasObject.weighted_bollinger_bands = weighted_bollinger_bands
PandasObject.rolling_weighted_mean = rolling_weighted_mean
PandasObject.supertrend = supertrend

PandasObject.sma = sma
PandasObject.wma = wma
PandasObject.ema = wma
PandasObject.hma = hma

PandasObject.zlsma = zlsma
PandasObject.zlwma = zlema
PandasObject.zlema = zlema
PandasObject.zlhma = zlhma
PandasObject.zlma = zlma



from kinetick.blotter import Blotter


class MainBlotter(Blotter):
    pass  # we just need the name


def build():
    blotter = MainBlotter()
    return blotter


# ===========================================
if __name__ == "__main__":
    build().run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Modified version from QTPyLib: Quantitative Trading Python Library
# https://github.com/ranaroussi/qtpylib
# Copyright 2016-2018 Ran Aroussi
#
# Modified by vin8tech
# Copyright 2019-2021 vin8tech, vinay patil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from pandas import concat as pd_concat
from kinetick.bot import Bot
from kinetick.enums import PositionType
from kinetick.models import Position


# =============================================


class Instrument(str):
    """A string subclass that provides easy access to misc
    symbol-related methods and information.
    """

    parent = None
    tick_window = None
    bar_window = None

    _position = None

    bot = Bot()
    logger = logging.getLogger(__name__)

    # ---------------------------------------
    def _set_parent(self, parent):
        """ sets the parent object to communicate with """
        self.parent = parent

    # ---------------------------------------
    def _set_windows(self, ticks, bars):
        """ be aware of default windows """
        self.tick_window = ticks
        self.bar_window = bars

    # ---------------------------------------
    @staticmethod
    def _get_symbol_dataframe(df, symbol):
        try:
            # this produce a "IndexingError using Boolean Indexing" (on rare occasions)
            return df[(df['symbol'] == symbol) | (df['symbol_group'] == symbol)].copy()
        except Exception as e:
            df = pd_concat([df[df['symbol'] == symbol]], sort=True)
            df.loc[:, '_idx_'] = df.index
            return df.drop_duplicates(subset=['_idx_'], keep='last').drop('_idx_', axis=1)

    # ---------------------------------------
    def get_bars(self, lookback=None, as_dict=False):
        """ Get bars for this instrument

        :Parameters:
            lookback : int
                Max number of bars to get (None = all available bars)
            as_dict : bool
                Return a dict or a pd.DataFrame object

        :Retruns:
            bars : pd.DataFrame / dict
                The bars for this instruments
        """
        bars = self._get_symbol_dataframe(self.parent.bars, self)

        # add signal history to bars
        bars = self.parent._add_signal_history(df=bars, symbol=self)

        lookback = self.bar_window if lookback is None else lookback
        bars = bars[-lookback:]
        # if lookback is not None:
        #     bars = bars[-lookback:]

        if not bars.empty > 0:  # and bars['asset_class'].values[-1] not in ("OPT", "FOP")
            bars = bars.drop(bars.columns[
                          bars.columns.str.startswith('opt_')].tolist(), axis=1)

        if as_dict:
            bars.loc[:, 'datetime'] = bars.index
            bars = bars.to_dict(orient='records')
            if lookback == 1:
                bars = None if not bars else bars[0]

        return bars

    # ---------------------------------------
    def get_bar(self):
        """ Shortcut to self.get_bars(lookback=1, as_dict=True) """
        return self.get_bars(lookback=1, as_dict=True)

    # ---------------------------------------
    def get_ticks(self, lookback=None, as_dict=False):
        """ Get ticks for this instrument

        :Parameters:
            lookback : int
                Max number of ticks to get (None = all available ticks)
            as_dict : bool
                Return a dict or a pd.DataFrame object

        :Retruns:
            ticks : pd.DataFrame / dict
                The ticks for this instruments
        """
        ticks = self._get_symbol_dataframe(self.parent.ticks, self)

        lookback = self.tick_window if lookback is None else lookback
        ticks = ticks[-lookback:]
        # if lookback is not None:
        #     ticks = ticks[-lookback:]

        if not ticks.empty and ticks['asset_class'].values[-1] not in ("OPT", "FOP"):
            ticks.drop(ticks.columns[
                           ticks.columns.str.startswith('opt_')].tolist(),
                       inplace=True, axis=1)

        if as_dict:
            ticks.loc[:, 'datetime'] = ticks.index
            ticks = ticks.to_dict(orient='records')
            if lookback == 1:
                ticks = None if not ticks else ticks[0]

        return ticks

    # ---------------------------------------
    def get_tick(self):
        """ Shortcut to self.get_ticks(lookback=1, as_dict=True) """
        return self.get_ticks(lookback=1, as_dict=True)

    # ---------------------------------------
    def get_price(self):
        """ Shortcut to self.get_ticks(lookback=1, as_dict=True)['last'] """
        tick = self.get_ticks(lookback=1, as_dict=True)
        return None if tick is None else tick['last']

    # ---------------------------------------
    def get_quote(self):
        """ Get last quote for this instrument

        :Retruns:
            quote : dict
                The quote for this instruments
        """
        if self in self.parent.quotes.keys():
            return self.parent.quotes[self]
        return None

    # ---------------------------------------
    def get_orderbook(self):
        """Get orderbook for the instrument

        :Retruns:
            orderbook : dict
                orderbook dict for the instrument
        """
        if self in self.parent.books.keys():
            return self.parent.books[self]

        return {
            "bid": [0], "bidsize": [0],
            "ask": [0], "asksize": [0]
        }

    # ---------------------------------------
    def create_position(self, entry_price, stop_loss, quantity=None, pos_type=PositionType.CO) -> Position:
        """
        return trade if all the conditions are met
        :param pos_type: position variety. ex MIS, CO, CNC. default to MIS. Possible types are defined in enums.PositionType
        :param quantity: quantity will be calculated based on risk assessment if null.
        :param entry_price:
        :param stop_loss:
        :return: position:Position
        """
        if self.parent.risk_assessor is not None:
            position = self.parent.risk_assessor.create_position(entry_price, stop_loss, quantity=quantity)
            position._tickerId = str(self.get_tickerId())
            position._symbol = self
            position.algo = self.parent.name
            position._variety = pos_type
            return position
        else:
            direction = "LONG" if entry_price > stop_loss else "SHORT"
            return Position(_tickerId=str(self.get_tickerId()), _symbol=self, entry_price=entry_price,
                            stop=stop_loss, _direction=direction, algo=self.parent.name, _quantity=quantity,
                            _variety=pos_type)

    # ---------------------------------------
    def open_position(self, position: Position, **kwargs):
        if position.active or self._position is not None:
            raise Exception("Position can't be opened because there is an active open position")

        self._position = position

        txn_type = position.direction.replace("LONG", "BUY").replace("SHORT", "SELL")
        if self.parent.backtest:
            self.order(txn_type, position.quantity, **kwargs)
        else:
            def callback(trade=position, txn=txn_type, commands=(), opts=kwargs, **args):
                cancel = True if 'cancel' in commands else False
                market = True if 'market' in commands else False
                if cancel:
                    self._position = None
                    self.logger.info(f'Order cancelled - ${trade.symbol}')
                else:
                    position.open_position()
                    self.parent.risk_assessor.enter_position(position)
                    order_id = self.order(txn, trade.quantity, limit_price=0 if market else trade.entry_price,
                                          # target=trade.target, providing
                                          # target will result in BO order
                                          pos_type=trade.variety,
                                          initial_stop=trade.stop, **opts)
                    position._broker_order_id = order_id
                    self.save_to_db(position)

            self.bot.send_order(position, "**ENTER** #" + self, callback=callback,
                                commands=('limit', 'cancel', 'market'))

    # ---------------------------------------
    def close_position(self, position: Position, exit_price=0, force=False, **kwargs):
        if self._position is not None and position is self._position:
            self._position = None
        if not position.active:
            raise Exception("Position can't be closed because the status is inactive")

        txn_type = "SELL" if position.direction == "LONG" else "BUY"  # EXIT
        tick = self.get_tick()
        exit_price = exit_price or (tick['last'] if tick else
                                    self.get_bar()['close'] if self.get_bar() else 0)
        position.exit_price = exit_price
        position.close_position()

        def _close(trade=position, txn=txn_type, market=True, opts=kwargs, **args):
            if trade.variety == PositionType.MIS or trade.variety == PositionType.CNC:
                # TODO if MIS verify if position is open with broker executed.
                self.order(txn, trade.quantity, pos_type=trade.variety,
                           limit_price=0 if market else trade.exit_price,
                           **opts, **args)
            else:
                self.cancel_order(trade.broker_order_id)

            self.parent.risk_assessor.exit_position(position)

        if self.parent.backtest:
            self.order(txn_type, position.quantity, **kwargs)
        elif force:
            _close()
        else:
            def callback(trade=position, commands=(), **args):
                market = True if 'market' in commands else False
                cancel = True if 'cancel' in commands else False
                if cancel:
                    trade.open_position()
                    self._position = trade
                    self.save_to_db(trade)
                else:
                    _close(market=market, **args)

            self.bot.send_order(position, "**EXIT** #" + self, callback=callback,
                                commands=('limit', 'cancel', 'market'))

        self.save_to_db(position)

    # ---------------------------------------
    def order(self, txn_type, quantity, **kwargs):
        """ Send an order for this instrument

        :Parameters:

            direction : string
                Order Type (BUY/SELL, EXIT/FLATTEN)
            quantity : int
                Order quantity

        :Optional:

            limit_price : float
                In case of a LIMIT order, this is the LIMIT PRICE
            expiry : int
                Cancel this order if not filled after *n* seconds (default 60 seconds)
            order_type : string
                Type of order: Market (default), LIMIT (default when limit_price is passed),
                MODIFY (required passing or orderId)
            orderId : int
                If modifying an order, the order id of the modified order
            target : float
                target (exit) price
            initial_stop : float
                price to set hard stop
            stop_limit: bool
                Flag to indicate if the stop should be STOP or STOP LIMIT (default False=STOP)
            trail_stop_at : float
                price at which to start trailing the stop
            trail_stop_by : float
                % of trailing stop distance from current price
            fillorkill: bool
                fill entire quantiry or none at all
            iceberg: bool
                is this an iceberg (hidden) order
            tif: str
                time in force (DAY, GTC, IOC, GTD). default is ``DAY``
        """
        txn_type = txn_type.upper().replace("LONG", "BUY").replace("SHORT", "SELL")
        return self.parent.order(txn_type.upper(), self, quantity, **kwargs)

    # ---------------------------------------
    def cancel_order(self, order_id):
        """ Cancels an order for this instrument

        :Parameters:
            orderId : int
                Order ID
        """

        return self.parent.cancel_order(order_id)

    # ---------------------------------------
    def market_order(self, direction, quantity, **kwargs):
        """ Shortcut for ``instrument.order(...)`` and accepts all of its
        `optional parameters <#qtpylib.instrument.Instrument.order>`_

        :Parameters:
            direction : string
                Order Type (BUY/SELL, EXIT/FLATTEN)
            quantity : int
                Order quantity
        """
        kwargs['limit_price'] = 0
        kwargs['order_type'] = "MARKET"
        self.parent.order(direction.upper(), self, quantity=quantity, **kwargs)

    # ---------------------------------------
    def limit_order(self, direction, quantity, price, **kwargs):
        """ Shortcut for ``instrument.order(...)`` and accepts all of its
        `optional parameters <#qtpylib.instrument.Instrument.order>`_

        :Parameters:
            direction : string
                Order Type (BUY/SELL, EXIT/FLATTEN)
            quantity : int
                Order quantity
            price : float
                Limit price
        """
        kwargs['limit_price'] = price
        kwargs['order_type'] = "LIMIT"
        self.parent.order(direction.upper(), self, quantity=quantity, **kwargs)

    # ---------------------------------------
    def buy(self, quantity, **kwargs):
        """ Shortcut for ``instrument.order("BUY", ...)`` and accepts all of its
        `optional parameters <#qtpylib.instrument.Instrument.order>`_

        :Parameters:
            quantity : int
                Order quantity
        """
        self.parent.order("BUY", self, quantity=quantity, **kwargs)

    # ---------------------------------------
    def buy_market(self, quantity, **kwargs):
        """ Shortcut for ``instrument.order("BUY", ...)`` and accepts all of its
        `optional parameters <#qtpylib.instrument.Instrument.order>`_

        :Parameters:
            quantity : int
                Order quantity
        """
        kwargs['limit_price'] = 0
        kwargs['order_type'] = "MARKET"
        self.parent.order("BUY", self, quantity=quantity, **kwargs)

    # ---------------------------------------
    def buy_limit(self, quantity, price, **kwargs):
        """ Shortcut for ``instrument.order("BUY", ...)`` and accepts all of its
        `optional parameters <#qtpylib.instrument.Instrument.order>`_

        :Parameters:
            quantity : int
                Order quantity
            price : float
                Limit price
        """
        kwargs['limit_price'] = price
        kwargs['order_type'] = "LIMIT"
        self.parent.order("BUY", self, quantity=quantity, **kwargs)

    # ---------------------------------------
    def sell(self, quantity, **kwargs):
        """ Shortcut for ``instrument.order("SELL", ...)`` and accepts all of its
        `optional parameters <#qtpylib.instrument.Instrument.order>`_

        :Parameters:
            quantity : int
                Order quantity
        """
        self.parent.order("SELL", self, quantity=quantity, **kwargs)

    # ---------------------------------------
    def sell_market(self, quantity, **kwargs):
        """ Shortcut for ``instrument.order("SELL", ...)`` and accepts all of its
        `optional parameters <#qtpylib.instrument.Instrument.order>`_

        :Parameters:
            quantity : int
                Order quantity
        """
        kwargs['limit_price'] = 0
        kwargs['order_type'] = "MARKET"
        self.parent.order("SELL", self, quantity=quantity, **kwargs)

    # ---------------------------------------
    def sell_limit(self, quantity, price, **kwargs):
        """ Shortcut for ``instrument.order("SELL", ...)`` and accepts all of its
        `optional parameters <#qtpylib.instrument.Instrument.order>`_

        :Parameters:
            quantity : int
                Order quantity
            price : float
                Limit price
        """
        kwargs['limit_price'] = price
        kwargs['order_type'] = "LIMIT"
        self.parent.order("SELL", self, quantity=quantity, **kwargs)

    # ---------------------------------------
    def exit(self):
        """ Shortcut for ``instrument.order("EXIT", ...)``
        (accepts no parameters)"""
        self.parent.order("EXIT", self)

    # ---------------------------------------
    def flatten(self):
        """ Shortcut for ``instrument.order("FLATTEN", ...)``
        (accepts no parameters)"""
        self.parent.order("FLATTEN", self)

    def clear(self):
        if self._position and self._position.active:
            self.logger.warning("Called clear when there's an active position")
        self._position = None
        self.logger.warning("Reset performed on " + self)

    # ---------------------------------------
    def get_contract(self):
        """Get contract object for this instrument

        :Retruns:
            contract : Object
                IB Contract object
        """
        return self.parent.get_contract(self)

    # ---------------------------------------
    def get_contract_details(self):
        """Get contract details for this instrument

        :Retruns:
            contract_details : dict
                IB Contract details
        """
        return self.parent.get_contract_details(self)

    # ---------------------------------------
    def get_tickerId(self) -> object:
        """Get contract's tickerId for this instrument

        :Retruns:
            tickerId : int
                IB Contract's tickerId
        """
        return self.parent.get_tickerId(self)

    # ---------------------------------------
    def get_combo(self):
        """Get instrument's group if part of an instrument group
        !IMPORTANT: NOT SUPPORTED. UNSTABLE API
        """
        raise Exception("Not supported")

    # ---------------------------------------
    def get_positions(self, attr=None):
        """Get the positions data for the instrument
        !IMPORTANT: NOT SUPPORTED. UNSTABLE API
        :Optional:
            attr : string
                Position attribute to get
                (optional attributes: symbol, position, avgCost, account)

        :Retruns:
            positions : dict (positions) / float/str (attribute)
                positions data for the instrument
        """
        pos = [self._position] if self._position is not None else []
        return pos

    # ---------------------------------------
    def set_position(self, position):
        self._position = position

    # ---------------------------------------
    def get_portfolio(self):
        """Get portfolio data for the instrument
        !IMPORTANT: NOT SUPPORTED. UNSTABLE API
        :Retruns:
            portfolio : dict
                portfolio data for the instrument
        """
        raise Exception("Not supported")

    # ---------------------------------------
    def get_orders(self):
        """Get orders for the instrument
        !IMPORTANT: NOT SUPPORTED. UNSTABLE API
        :Retruns:
            orders : list
                list of order data as dict
        """
        raise Exception("Not supported")

    # ---------------------------------------
    def get_pending_orders(self):
        """Get pending orders for the instrument
        !IMPORTANT: NOT SUPPORTED. UNSTABLE API
        :Retruns:
            orders : list
                list of pending order data as dict
        """
        raise Exception("Not supported")

    # ---------------------------------------
    def get_active_order(self, order_type="STOP"):
        """Get artive order id for the instrument by order_type
        !IMPORTANT: NOT SUPPORTED. UNSTABLE API
        :Optional:
            order_type : string
                the type order to return: STOP (default), LIMIT, MARKET

        :Retruns:
            order : object
                IB Order object of instrument
        """
        raise Exception("Not supported")

    # ---------------------------------------
    def get_trades(self):
        """Get orderbook for the instrument
        !IMPORTANT: NOT SUPPORTED. UNSTABLE API
        :returns:
            trades : pd.DataFrame
                instrument's trade log as DataFrame
        """
        raise Exception("Not supported")

    # ---------------------------------------
    def get_symbol(self):
        """Get symbol of this instrument

        :Retruns:
            symbol : string
                instrument's symbol
        """
        return self

    # ---------------------------------------
    def modify_order(self, orderId, quantity=None, limit_price=None):
        """Modify quantity and/or limit price of an active order for the instrument
        !IMPORTANT: NOT SUPPORTED. UNSTABLE API
        :Parameters:
            orderId : int
                the order id to modify

        :Optional:
            quantity : int
                the required quantity of the modified order
            limit_price : int
                the new limit price of the modified order
        """
        raise Exception("Not supported")

    # ---------------------------------------
    def modify_order_group(self, orderId, entry=None, target=None, stop=None, quantity=None):
        """Modify bracket order
        !IMPORTANT: NOT SUPPORTED. UNSTABLE API
        :Parameters:
            orderId : int
                the order id to modify

        :Optional:
            entry: float
                new entry limit price (for unfilled limit orders only)
            target: float
                new target limit price (for unfilled limit orders only)
            stop: float
                new stop limit price (for unfilled limit orders only)
            quantity : int
                the required quantity of the modified order
        """
        raise Exception("Not supported")

    # ---------------------------------------
    def move_stoploss(self, stoploss):
        """Modify stop order.
        Auto-discover **orderId** and **quantity** and invokes ``self.modify_order(...)``.
        !IMPORTANT: NOT SUPPORTED. UNSTABLE API
        :Parameters:
            stoploss : float
                the new stoploss limit price

        """
        raise Exception("Not supported")

    # ---------------------------------------
    def get_margin_requirement(self):
        """ Get margin requirements for intrument (futures only)
        !IMPORTANT: NOT SUPPORTED. UNSTABLE API
        :Retruns:
            margin : dict
                margin requirements for instrument
                (all values are ``None`` for non-futures instruments)
        """
        raise Exception("Not supported yet!")

    # ---------------------------------------
    def get_max_contracts_allowed(self, overnight=True):
        """ Get maximum contracts allowed to trade
        baed on required margin per contract and
        current account balance (futures only)
        !IMPORTANT: NOT SUPPORTED. UNSTABLE API
        :Parameters:
            overnight : bool
                Calculate based on Overnight margin (set to ``False`` to use Intraday margin req.)

        :Retruns:
            contracts : int
                maximum contracts allowed to trade (returns ``None`` for non-futures)
        """
        raise Exception("Not supported yet!")

    def get_margin_max_contracts(self, overnight=True):
        """ Deprecated (renamed to ``get_max_contracts_allowed``)"""
        raise Exception("Not supported yet!")

    # ---------------------------------------
    def get_ticksize(self):
        """ Get instrument ticksize

        :Retruns:
            ticksize : int
                Min. tick size
        """
        ticksize = self.parent.get_contract_details(self)['m_minTick']
        return float(ticksize)

    # ---------------------------------------
    def pnl_in_range(self, min_pnl, max_pnl):
        """ Check if instrument pnl is within given range
        !IMPORTANT: NOT SUPPORTED. UNSTABLE API
        :Parameters:
            min_pnl : flaot
                minimum session pnl (in USD / IB currency)
            max_pnl : flaot
                maximum session pnl (in USD / IB currency)

        :Retruns:
            status : bool
                if pnl is within range
        """
        portfolio = self.get_portfolio()
        return -abs(min_pnl) < portfolio['totalPNL'] < abs(max_pnl)

    # ---------------------------------------
    def log_signal(self, signal):
        """ Log Signal for instrument

        :Parameters:
            signal : integer
                signal identifier (1, 0, -1)
        """
        return self.parent._log_signal(self, signal)

    # ---------------------------------------
    def save_to_db(self, obj):
        if not self.parent.blotter_args["dbskip"]:
            obj.save()

    # ---------------------------------------
    @property
    def bars(self):
        """(Property) Shortcut to self.get_bars()"""
        return self.get_bars()

    # ---------------------------------------
    @property
    def bar(self):
        """(Property) Shortcut to self.get_bar()"""
        return self.get_bar()

    # ---------------------------------------
    @property
    def ticks(self):
        """(Property) Shortcut to self.get_ticks()"""
        return self.get_ticks()

    # ---------------------------------------
    @property
    def tick(self):
        """(Property) Shortcut to self.get_tick()"""
        return self.get_tick()

    # ---------------------------------------
    @property
    def price(self):
        """(Property) Shortcut to self.get_price()"""
        return self.get_price()

    # ---------------------------------------
    @property
    def quote(self):
        """(Property) Shortcut to self.get_quote()"""
        return self.get_quote()

    # ---------------------------------------
    @property
    def orderbook(self):
        """(Property) Shortcut to self.get_orderbook()"""
        return self.get_orderbook()

    # ---------------------------------------
    @property
    def symbol(self):
        """(Property) Shortcut to self.get_symbol()"""
        return self

    # ---------------------------------------
    @property
    def contract(self):
        """(Property) Shortcut to self.get_contract()"""
        return self.get_contract()

    # ---------------------------------------
    @property
    def contract_details(self):
        """(Property) Shortcut to self.get_contract_details()"""
        return self.get_contract_details()

    # ---------------------------------------
    @property
    def tickerId(self):
        """(Property) Shortcut to self.get_tickerId()"""
        return self.get_tickerId()

    # ---------------------------------------
    @property
    def combo(self):
        """(Property) Shortcut to self.get_combo()"""
        return self.get_combo()

    # ---------------------------------------
    @property
    def positions(self):
        """(Property) Shortcut to self.get_positions()"""
        raise Exception("Not Supported")

    # ---------------------------------------
    @property
    def position(self):
        """(Property) Shortcut to self._position"""
        return self._position

    # ---------------------------------------
    @property
    def portfolio(self):
        """(Property) Shortcut to self.get_portfolio()"""
        raise Exception("Not Supported")

    # ---------------------------------------
    @property
    def orders(self):
        """(Property) Shortcut to self.get_orders()"""
        return self.get_orders()

    # ---------------------------------------
    @property
    def pending_orders(self):
        """(Property) Shortcut to self.get_pending_orders()"""
        return self.get_pending_orders()

    # ---------------------------------------
    @property
    def trades(self):
        """(Property) Shortcut to self.get_trades()"""
        return self.get_trades()

    # ---------------------------------------
    @property
    def margin_requirement(self):
        """(Property) Shortcut to self.get_margin_requirement()"""
        return self.get_margin_requirement()

    # ---------------------------------------
    @property
    def margin_max_contracts(self):
        """ Deprecated (renamed to ``max_contracts_allowed``)"""
        return self.get_max_contracts_allowed()

    @property
    def max_contracts_allowed(self):
        """(Property) Shortcut to self.get_max_contracts_allowed()"""
        return self.get_max_contracts_allowed()

    # ---------------------------------------
    @property
    def ticksize(self):
        """(Property) Shortcut to self.get_ticksize()"""
        return self.get_ticksize()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# QTPyLib: Quantitative Trading Python Library
# https://github.com/ranaroussi/qtpylib
#
# Copyright 2016-2018 Ran Aroussi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from threading import Thread, Semaphore
from multiprocessing import Process, cpu_count
from sys import exit as sysexit, version_info as sys_version_info
from os import _exit as osexit
from time import sleep, time

# =============================================
# check min, python version
if sys_version_info < (3, 4):
    raise SystemError("QTPyLib requires Python version >= 3.4")
# =============================================


class multitasking():
    """
    Non-blocking Python methods using decorators
    (a class-based implementation of the multitasking library)
    https://github.com/ranaroussi/multitasking
    """

    __KILL_RECEIVED__ = False
    __TASKS__ = []

    # processing
    __CPU_CORES__ = cpu_count()
    __POOLS__ = {}
    __POOL_NAME__ = "Main"

    @classmethod
    def getPool(cls, name=None):
        if name is None:
            name = cls.__POOL_NAME__

        return {
            "engine": "thread" if cls.__POOLS__[cls.__POOL_NAME__
                                                ]["engine"] == Thread else "process",
            "name": name,
            "threads": cls.__POOLS__[cls.__POOL_NAME__]["threads"]
        }

    @classmethod
    def createPool(cls, name="main", threads=None, engine="thread"):

        cls.__POOL_NAME__ = name

        # if threads is None:
        #     threads = cls.__CPU_CORES__

        try:
            threads = int(threads)
        except Exception as e:
            threads = 1

        # 1 thread is no threads
        if threads < 2:
            threads = 0

        cls.__POOLS__[cls.__POOL_NAME__] = {
            "pool": Semaphore(threads) if threads > 0 else 1,
            "engine": Process if "process" in engine.lower() else Thread,
            "name": name,
            "threads": threads
        }

    @classmethod
    def task(cls, callee):

        # create default pool if nont exists
        if not cls.__POOLS__:
            cls.createPool()

        def _run_via_pool(*args, **kwargs):
            with cls.__POOLS__[cls.__POOL_NAME__]['pool']:
                return callee(*args, **kwargs)

        def async_method(*args, **kwargs):
            # no threads
            if cls.__POOLS__[cls.__POOL_NAME__]['threads'] == 0:
                return callee(*args, **kwargs)

            # has threads
            if not cls.__KILL_RECEIVED__:
                task = cls.__POOLS__[cls.__POOL_NAME__]['engine'](
                    target=_run_via_pool, args=args, kwargs=kwargs, daemon=False)
                cls.__TASKS__.append(task)
                task.start()
                return task

            return None

        return async_method

    @classmethod
    def wait_for_tasks(cls):
        cls.__KILL_RECEIVED__ = True

        if cls.__POOLS__[cls.__POOL_NAME__]['threads'] == 0:
            return True

        try:
            running = len([t.join(1)
                           for t in cls.__TASKS__ if t is not None and t.isAlive()])
            while running > 0:
                running = len(
                    [t.join(1) for t in cls.__TASKS__ if t is not None and t.isAlive()])
        except Exception as e:
            pass
        return True

    @classmethod
    def killall(cls):
        cls.__KILL_RECEIVED__ = True
        try:
            sysexit(0)
        except SystemExit:
            osexit(0)

# =============================================


class RecurringTask(Thread):
    """Calls a function at a sepecified interval."""

    def __init__(self, func, interval_sec, init_sec, *args, **kwargs):
        """Call `func` every `interval_sec` seconds.

        Starts the timer.

        Accounts for the runtime of `func` to make intervals as close to `interval_sec` as possible.
        args and kwargs are passed to Thread().

        :Parameters:
            func : object
                Function to invoke every N seconds
            interval_sec : int
                Call func every this many seconds
            init_sec : int
                Wait this many seconds initially before the first call
            *args : mixed
                parameters sent to parent Thread class
            **kwargs : mixed
                parameters sent to parent Thread class
        """

        # threading.Thread.__init__(self, *args, **kwargs) # For some reason super() doesn't work
        super().__init__(*args, **kwargs)  # Works!
        self._func = func
        self.interval_sec = interval_sec
        self.init_sec = init_sec
        self._running = True
        self._functime = None  # Time the next call should be made

        self.start()

    def __repr__(self):
        return 'RecurringTask({}, {}, {})'.format(self._func, self.interval_sec, self.init_sec)

    def run(self):
        """Start the recurring task."""
        if self.init_sec:
            sleep(self.init_sec)
        self._functime = time()
        while self._running:
            start = time()
            self._func()
            self._functime += self.interval_sec
            if self._functime - start > 0:
                sleep(self._functime - start)

    def stop(self):
        """Stop the recurring task."""
        self._running = False



# Copyright 2021 vin8tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging
import os

from kinetick.utils import utils

from kinetick.algo import Algo
from kinetick.bot import Bot
from kinetick.enums import Timeframes
from kinetick.risk_assessor import RiskAssessor
import pandas as pd
import numpy as np
from decimal import getcontext

from kinetick.lib.indicators.talib_indicators import MA

getcontext().prec = 6

utils.create_logger(__name__, level=os.getenv('LOGLEVEL') or logging.INFO)
logger = logging.getLogger(__name__)

__LOOKBACK__ = 800


class MacdSuperStrategy(Algo):
    """
    slow fast macd cross over
    entry on supertrend stop at supertrend and VWAP
    stochastic
    """

    count = {}

    ta_df = pd.DataFrame({'open': [0], 'high': [0], 'low': [0], 'close': [0], 'vwap': [0], 'ma': [0],
                          'macd': [0], 'macdsignal': [0], 'macdhist': [0], 'ATR_7': [0], 'ST': [0], 'STX': [0],
                          'slow_k': [0], 'slow_d': [0], 'signal': [0]})

    ti = {}  # symbol:ta
    _bars = {"~": pd.DataFrame()}

    def __init__(self, instruments, fast_period=1, slow_period=15, **kwargs):
        super().__init__(instruments, **kwargs)
        self.__MIN_PERIOD__ = Timeframes.timeframe_to_minutes(Timeframes.to_timeframe(self.resolution))
        self.__FAST_PERIOD__ = fast_period
        self.__SLOW_PERIOD__ = slow_period
        self.bot = Bot()

    # ---------------------------------------
    def gen_signal(self, macd_hist, stx, k, d):
        if utils.to_decimal(macd_hist, 2) > 0.0 and stx == 'up' \
                and k > d and (int(k) <= 90 or int(d) <= 90):
            return 'B'
        elif utils.to_decimal(macd_hist, 2) < 0.0 and stx == 'down' \
                and k < d and (k > 20 or d > 20):
            return 'S'
        return 'N'

    # ---------------------------------------
    def gen_slow_signal(self, hist, k, d):
        if hist > 0.0 and (k <= 90 or d <= 90):
            return 'B'
        elif hist < 0.0 and (k >= 30 or d >= 30):
            return 'S'
        return 'N'

    # ---------------------------------------
    def fill_ta(self, **kwargs):
        df = pd.DataFrame()
        for key, value in kwargs.items():
            df[key] = value[-2:]
        return df

    # ---------------------------------------
    def gen_indicators(self, bars, fast=True):
        ma = MA(data=bars, timeperiod=5)  # TODO maybe redundant on slow ti
        macd = bars['close'].macd(fast=12, slow=26, smooth=9)
        stoch = bars.stoch(window=26, d=6, k=6, fast=False)
        if fast:
            super_trend = bars.supertrend(period=7, multiplier=3)
        else:
            super_trend = pd.DataFrame({'ATR_7': [0], 'ST': [0], 'STX': [0]})
            # stoch = pd.DataFrame({'slow_k': [0], 'slow_d': [0]})
        return ma, macd, super_trend, stoch

    # ---------------------------------------
    def get_statistics(self, instrument, period=1, bars=None):
        if bars is None:
            bars = self.sync_bars(instrument)

        if 'vwap' not in bars.columns:
            bars['vwap'] = bars.vwap()

        if not period == self.__MIN_PERIOD__:
            ticks = str(period) + 'T'
            bars = bars.resample(ticks).agg({
                'open': 'first',
                'high': 'max',
                'low': 'min',
                'close': 'last',
                'vwap': 'last',
                # vwap could be missing initially until sync_bars is called and therefore
                # the resulting bars would be empty
                'volume': 'sum'}).dropna()

        fast = False if not period == self.__FAST_PERIOD__ else True

        ma, macd, super_trend, stoch = self.gen_indicators(bars, fast)

        params = {
            'open': bars['open'],
            'high': bars['high'],
            'low': bars['low'],
            'close': bars['close'],
            'volume': bars['volume'],
            'ma': ma,
            'macd': macd['macd'],
            'macdsignal': macd['signal'],
            'macdhist': macd['histogram'],
            'ATR_7': super_trend['ATR_7'],
            'ST': super_trend['ST'],
            'STX': super_trend['STX'],
            'slow_k': stoch['slow_k'],
            'slow_d': stoch['slow_d']
        }

        if 'vwap' in bars.columns:
            params['vwap'] = bars[['vwap']].fillna(method='pad')

        return self.fill_ta(**params)

    # ---------------------------------------
    def fill_slow_ta(self, instrument):
        bars = instrument.get_bars(lookback=__LOOKBACK__)
        # bars = self._bars[instrument]
        slow_ta = self.get_statistics(instrument=instrument, period=self.__SLOW_PERIOD__, bars=bars)
        slow_ta['signal'] = \
            np.vectorize(self.gen_slow_signal)(slow_ta['macdhist'], slow_ta['slow_k'], slow_ta['slow_d'])
        self.ti[instrument]['slow_ta'] = slow_ta
        with pd.option_context('display.max_rows', None, 'display.max_columns', None):
            logger.debug("--------\nSLOW TA - %s\n %s", instrument, self.ti[instrument]['slow_ta'][-1:])

    # ---------------------------------------
    def fill_fast_ta(self, instrument):
        bars = instrument.get_bars(lookback=__LOOKBACK__)
        # bars = self._bars[instrument]  # append raw ticks from time delta
        fast_ta = self.get_statistics(instrument, period=self.__FAST_PERIOD__, bars=bars)
        fast_ta['signal'] = \
            np.vectorize(self.gen_signal)(fast_ta['macdhist'], fast_ta['STX'], fast_ta['slow_k'], fast_ta['slow_d'])
        self.ti[instrument]['fast_ta'] = fast_ta
        with pd.option_context('display.max_rows', None, 'display.max_columns', None):
            logger.debug("--------\nFAST TA - %s\n %s", instrument, self.ti[instrument]['fast_ta'][-1:])

    # ---------------------------------------
    def exit_trade(self, instrument):
        try:
            logger.info(f'Exit position {instrument}')
            position = instrument.position
            instrument.close_position(position)

            self.record(time=position.exit_time)
            self.record(EXIT_REASON=position.exit_reason)
            return position
        except Exception as e:
            logger.error("Error exiting trade %s", e)

    # ---------------------------------------
    def enter_trade(self, instrument, entry_price, stop_loss):
        try:
            logger.info(f'Enter position {instrument}')
            position = instrument.create_position(entry_price, stop_loss)

            instrument.open_position(position)
            return position
        except Exception as e:
            logger.error("Error entering trade %s", e)

    # ---------------------------------------

    def init_instruments(self, instruments=None):
        if instruments is None:
            instruments = self.instruments
        for instrument in instruments:
            self.count[instrument] = 0
            if instrument not in self.ti.keys():
                self.ti[instrument] = {'fast_ta': self.ta_df.copy(), 'slow_ta': self.ta_df.copy()}

    # ---------------------------------------
    def add_instrument(self, symbol):
        pass

    # ---------------------------------------
    def resetHandler(self, update, context):
        self.risk_assessor.reset()
        self.log_algo.info("Strategy reset complete.")
        update.message.reply_text("Strategy reset complete.")

    # ---------------------------------------
    def health(self, update, context):
        update.message.reply_text(f'count={self.count}')

    # ---------------------------------------

    def on_start(self):
        """ initialize tick counter """
        self.init_instruments()

        self.bot.add_command_handler("reset", self.resetHandler, "Reset")
        self.bot.add_command_handler("health", self.health, "Health Check")

    # ---------------------------------------
    def on_quote(self, instrument):
        pass

    # ---------------------------------------
    def on_orderbook(self, instrument):
        pass

    # ---------------------------------------
    def on_fill(self, instrument, order):
        pass

    # ----------------------------------------
    def apply(self, instrument, ltp, timestamp=None):
        # self.fill_fast_ta(instrument)
        slow_ta = self.ti[instrument]['slow_ta'].to_dict(orient='records')
        fast_ta = self.ti[instrument]['fast_ta'].to_dict(orient='records')

        # validate stop-loss and target hit
        open_positions = [instrument.position] if instrument.position is not None else []
        for position in open_positions:
            """Exit checks"""
            direction = position.direction

            sl_hit = False
            if direction == 'LONG':
                sl_hit = True if ltp <= position.stop else False
            elif direction == 'SHORT':
                sl_hit = True if ltp >= position.stop else False

            if (ltp <= position.target and direction == "SHORT") or \
                    (ltp >= position.target and direction == "LONG") or sl_hit:
                """target or stop-loss hit (exit conditions)"""
                position.exit_price = ltp
                position.exit_time = timestamp

                if sl_hit:
                    position.exit_reason = 'SL Hit'
                    self.exit_trade(instrument)

                elif (fast_ta[-1]['signal'] == 'B' or (fast_ta[-1]['slow_k'] > fast_ta[-1]['slow_d'])) \
                        and position.direction == "SHORT":
                    position.exit_reason = 'Target Hit'
                    self.exit_trade(instrument)

                elif (fast_ta[-1]['signal'] == 'S' or (fast_ta[-1]['slow_k'] < fast_ta[-1]['slow_d'])) \
                        and position.direction == "LONG":
                    position.exit_reason = 'Target Hit'
                    self.exit_trade(instrument)
                else:
                    # TODO trail stop-loss
                    position.exit_reason = 'Target Hit'
                    self.exit_trade(instrument)

        """ entry conditions """
        if len(open_positions) == 0:

            entry_price = (fast_ta[-1]['ma'] + ltp) / 2

            if slow_ta[-1]['signal'] == 'B' and fast_ta[-1]['signal'] == 'B':
                # one minute super trend
                bars = instrument.get_bars(lookback=__LOOKBACK__)
                ma, macd, super_trend, stoch = self.gen_indicators(bars, fast=True)
                if super_trend.iloc[-1]['STX'] == 'up':
                    entry_price = entry_price if entry_price < ltp else ltp
                    entry_price = 5 * round(entry_price / 5, 2)
                    st = fast_ta[-1]['ST']
                    st_buffer = st * 0.001
                    logger.info(f'BUY {instrument} at: {entry_price}')
                    logger.info("FAST TA - %s\n %s", instrument, json.dumps(fast_ta[-1], indent=2))
                    logger.info("SLOW TA - %s\n %s", instrument, json.dumps(slow_ta[-1], indent=2))
                    self.enter_trade(instrument, entry_price,
                                     stop_loss=utils.round_to_fraction(st - st_buffer, 0.05))

            elif slow_ta[-1]['signal'] == 'S' and fast_ta[-1]['signal'] == 'S':
                # one minute super trend
                bars = instrument.get_bars(lookback=__LOOKBACK__)
                ma, macd, super_trend, stoch = self.gen_indicators(bars, fast=True)
                if super_trend.iloc[-1]['STX'] == 'down':
                    entry_price = entry_price if entry_price > ltp else ltp
                    entry_price = 5 * round(entry_price / 5, 2)
                    st = fast_ta[-1]['ST']
                    st_buffer = st * 0.001
                    logger.info(f'SELL {instrument} at: {entry_price}')
                    logger.info("FAST TA - %s\n %s", instrument, json.dumps(fast_ta[-1], indent=2))
                    logger.info("SLOW TA - %s\n %s", instrument, json.dumps(slow_ta[-1], indent=2))
                    self.enter_trade(instrument, entry_price,
                                     stop_loss=utils.round_to_fraction(st + st_buffer, 0.05))

    # ---------------------------------------
    def on_tick(self, instrument):
        # increase counter and do nothing if nor 100th tick
        self.count[instrument] += 1

        if self.count[instrument] % 100 != 0:
            return

        # get last tick dict
        tick = instrument.get_ticks(lookback=1, as_dict=True)
        time = tick['datetime']
        self.apply(instrument, tick['last'], time)

    # ---------------------------------------
    def on_bar(self, instrument):
        if self.backtest:
            self.simulate(instrument)
        else:
            self.fill_slow_ta(instrument)
            self.fill_fast_ta(instrument)
            bar = instrument.get_bars(lookback=1, as_dict=True)
            self.apply(instrument, bar['close'], bar['datetime'])

    # --------------------------------------------
    def simulate(self, instrument):
        bars = instrument.get_bars(lookback=__LOOKBACK__)
        self.count[instrument] += 1
        if len(bars) < __LOOKBACK__:
            return

        bar = instrument.get_bars(lookback=1, as_dict=True)

        self.fill_slow_ta(instrument)
        self.fill_fast_ta(instrument)

        self.apply(instrument, bar['close'], bar['datetime'])
    # ----------------------------------------


# ===========================================
if __name__ == "__main__":
    strategy = MacdSuperStrategy(
        instruments=["ACC"],
        resolution=Timeframes.MINUTE_1,
        tick_window=50,
        bar_window=__LOOKBACK__,
        preload="1D",
        backtest=True,
        start="2020-07-01 00:15:00",
        risk_assessor=RiskAssessor(max_trades=4, initial_capital=120000, initial_margin=1000,
                                   risk2reward=1.2, risk_per_trade=100),
        # backfill=True,
        timezone="Asia/Calcutta"
    )
    strategy.run()



# Copyright 2021 vin8tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os

from kinetick.utils import utils

from kinetick.algo import Algo
from kinetick.enums import Timeframes, PositionType
from kinetick.risk_assessor import RiskAssessor
from decimal import getcontext

getcontext().prec = 6

utils.create_logger(__name__, level=os.getenv('LOGLEVEL') or logging.INFO)
logger = logging.getLogger(__name__)

__LOOKBACK__ = 800


class BuyLowSellHigh(Algo):
    """
    buy low sell high algorithm
    """

    def on_start(self):
        """ initialize your strategy. Perform any operation that are required to be run before starting algo."""
        pass

    # ---------------------------------------
    def on_quote(self, instrument):
        logger.debug(instrument.quote)

    # ---------------------------------------
    def on_orderbook(self, instrument):
        logger.debug(instrument.orderbook)

    # ---------------------------------------
    def on_fill(self, instrument, order):
        """ Not supported currently """
        pass

    # ---------------------------------------
    def on_tick(self, instrument):
        # get last tick dict
        logger.debug(instrument.tick)
        # get last 10 ticks
        logger.debug(instrument.get_ticks(lookback=10, as_dict=True))
        # get all the ticks
        ticks = instrument.get_ticks()

    # ---------------------------------------
    def on_bar(self, instrument):
        bars = instrument.get_bars(lookback=3)
        if instrument.position is None:
            """ check if there any open positions on the instrument."""
            if len(bars) >= 2:  # at least 2 bars are needed
                if bars['close'][-1] < bars['close'][0]:
                    position = instrument.create_position(entry_price=bars['close'][-1],
                                                          stop_loss=bars['low'][-1],
                                                          quantity=1, pos_type=PositionType.MIS)
                    """ Create position instance with inputs. 
                        If quantity is not provided then it will be automatically calculated based on risk parameters.
                    """
                    instrument.open_position(position)
                    """ instrument.open_position() will send order request to bot and 
                        attach the position instance to instrument to indicate about open position.
                        when order request is accepted from bot the position becomes active.
                    """
        elif instrument.position.active:
            """ check IF the position was executed in bot """
            if len(bars) >= 2:
                if bars['close'][-1] > bars['close'][0]:
                    instrument.close_position(instrument.position)
                    """ close the position. """

    # --------------------------------------------


# ===========================================
if __name__ == "__main__":
    strategy = BuyLowSellHigh(
        instruments=["ACC"],
        resolution=Timeframes.MINUTE_1,
        tick_window=50,
        bar_window=__LOOKBACK__,
        preload="1D",
        backtest=True,
        start="2020-07-01 00:15:00",
        risk_assessor=RiskAssessor(max_trades=4, initial_capital=120000, initial_margin=1000,
                                   risk2reward=1.2, risk_per_trade=100),
        # backfill=True,
        timezone="Asia/Calcutta"
    )
    strategy.run()



from mongoengine import StringField, DateTimeField, IntField, FloatField, BooleanField, DynamicDocument
from datetime import datetime


# note: position / order / trade are used interchangeably through the app.
class Position(DynamicDocument):
    """ Position Data Model.
    holds information relating to either trade/order/position.
    """
    _tickerId = StringField(max_length=50, required=True, db_field="tickerId")
    _symbol = StringField(max_length=50, required=False, db_field="symbol")
    datetime = DateTimeField(required=True, default=datetime.utcnow)
    algo = StringField(max_length=100)
    _direction = StringField(max_length=20, choices=('LONG', 'SHORT'), db_field="direction")
    _quantity = IntField(default=0, db_field="quantity")
    entry_time = DateTimeField()
    exit_time = DateTimeField()
    exit_reason = StringField()
    order_type = StringField()  # LIMIT/MARKET
    _broker_order_id = StringField(db_field="broker_order_id")
    _variety = StringField(db_field="variety")
    market_price = FloatField()
    target = FloatField(default=0.0)
    stop = FloatField(default=0.0)
    entry_price = FloatField(default=0.0)
    exit_price = FloatField(default=0.0)
    realized_pnl = FloatField(default=0.0)
    _active = BooleanField(default=False, db_field="active")
    tradingsymbol = StringField(max_length=50, required=False)
    opt_strike = FloatField(required=False)
    opt_type = StringField(require=False),
    opt_expiry = StringField(required=False),
    sec_type = StringField(default='STK')  # TODO add enum
    underlying = StringField(required=False)

    meta = {
        'indexes': [
            {
                'fields': ['_active'],
                'sparse': True
            },
            {
                'fields': ['algo'],
                'sparse': True
            }
        ]
    }

    def open_position(self):
        if self._direction is None:
            raise Exception("no direction provided")
        if self._quantity is None:
            raise Exception("no quantity provided")
        if self.entry_time is None:
            self.entry_time = datetime.now()
        self._active = True

    def close_position(self):
        if self._direction is None:
            raise Exception("no direction provided")
        if self._quantity is None:
            raise Exception("no quantity provided")
        if self.exit_time is None:
            self.exit_time = datetime.now()
        self.realized_pnl = self.pnl()
        self._active = False

    def pnl(self):
        if not self.exit_price:
            raise Exception("Trade exit_price is not provided")

        pnl = abs(self.exit_price - self.entry_price)

        sl_hit = False
        if self.exit_price <= self.entry_price and self._direction == "LONG":
            sl_hit = True
        elif self.exit_price >= self.entry_price and self._direction == "SHORT":
            sl_hit = True

        pnl = -pnl if sl_hit else pnl
        pnl = pnl * self._quantity
        return round(pnl, 2)

    @property
    def active(self):
        return self._active

    @property
    def ticker_id(self):
        return self._tickerId

    @property
    def symbol(self):
        return self._symbol

    @property
    def direction(self):
        return self._direction

    @property
    def quantity(self):
        return self._quantity

    @property
    def broker_order_id(self):
        return self._broker_order_id

    @property
    def variety(self):
        return self._variety

    @staticmethod
    def find(algo, **query) -> list:
        return Position.objects(algo=algo, **query)
